from __future__ import absolute_import
from unittest import TestCase
import numpy as np
from numpy import array
from .main import labels2contours


class TestSelect_labels2Contours(TestCase):
    def assert_output(self, expected_output, result):
        self.assertEqual(len(expected_output), len(result))
        for i in range(len(expected_output)):
            result_external = result[i][0]
            print(result_external)
            np.testing.assert_array_equal(result_external, expected_output[i][0])

            # Test inner contours
            result_internal_list = result[i][1]
            expected_internal_list = expected_output[i][1]
            self.assertEqual(len(expected_internal_list), len(result_internal_list))
            for result_internal, expected_internal in zip(
                result_internal_list, expected_internal_list
            ):
                np.testing.assert_array_equal(result_internal, expected_internal)

    def test_happy_flow(self):
        data = array(
            [
                [0, 5, 5, 0, 0, 0],
                [0, 5, 5, 0, 0, 0],
                [1, 1, 1, 0, 0, 0],
                [0, 2, 0, 0, 0, 0],
                [0, 2, 2, 0, 0, 0],
                [0, 2, 2, 2, 0, 0],
                [0, 2, 2, 2, 2, 0],
            ]
        )

        expected_output = [
            [[[0, 2]], [[2, 2]]],  # 1
            [[[1, 3]], [[1, 6]], [[4, 6]]],  # 2
            [[[1, 0]], [[1, 1]], [[2, 1]], [[2, 0]]],  # 5
        ]

        result = labels2contours(data)

        self.assert_output(expected_output, result)

    def test_return_inner_no_inner(self):
        data = array(
            [
                [0, 5, 5, 0, 0, 0],
                [0, 5, 5, 0, 0, 0],
                [1, 1, 1, 0, 0, 0],
                [0, 2, 0, 0, 0, 0],
                [0, 2, 2, 0, 0, 0],
                [0, 2, 2, 2, 0, 0],
                [0, 2, 2, 2, 2, 0],
            ]
        )

        expected_output = [
            [[[[0, 2]], [[2, 2]]], []],  # 1
            [[[[1, 3]], [[1, 6]], [[4, 6]]], []],  # 2
            [[[[1, 0]], [[1, 1]], [[2, 1]], [[2, 0]]], []],  # 5
        ]

        result = labels2contours(data, return_inner_contours=True)
        self.assert_output(expected_output, result)

    def test_donut(self):
        data = array(
            [
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 2, 2, 2, 2, 0],
                [0, 2, 0, 0, 2, 0],
                [0, 2, 0, 0, 2, 0],
                [0, 2, 2, 2, 2, 0],
            ]
        )

        expected_output = [
            [
                # External contour
                [[[1, 3]], [[1, 6]], [[4, 6]], [[4, 3]]],
                # All the internal contours
                [
                    # Corners are interpolated as being diagonal
                    # hench the 8 points for something that looks like a square
                    [
                        [[1, 4]],
                        [[2, 3]],
                        [[3, 3]],
                        [[4, 4]],
                        [[4, 5]],
                        [[3, 6]],
                        [[2, 6]],
                        [[1, 5]],
                    ]
                ],
            ],
        ]

        result = labels2contours(data, return_inner_contours=True)
        self.assert_output(expected_output, result)

    def test_deprated_warning(self):
        data = array(
            [
                [0, 5, 5, 0, 0, 0],
                [0, 5, 5, 0, 0, 0],
                [1, 1, 1, 0, 0, 0],
                [0, 2, 0, 0, 0, 0],
                [0, 2, 2, 0, 0, 0],
                [0, 2, 2, 2, 0, 0],
                [0, 2, 2, 2, 2, 0],
            ]
        )

        with self.assertRaises(DeprecationWarning):
            labels2contours(data, simplify_error=0.1)



from typing import List, Union
import copy
from ...transformers import json2contours
from ...contour_operations import get_center
from ..translation import translation_json


def json_object_to_keep(
    json_object: dict, new_x: tuple, new_y: tuple, crop_mode: str = "remove_objects"
) -> bool:
    """
    Checks if json_object is inside the cropped image.
    json_object: (dict) following the standard cytosmart format
    new_x (tuple) (xmin_crop, xmax_crop) the x dimensions of the cropped image
    new_y (tuple) (ymin_crop, ymax_crop) the y dimensions of the cropped images
    crop_mode: (string) "remove_objects" & "keep_objects". Default mode is 'remove_objects'
                using this mode removes json_object if center is outside the crop. "keep_objects"
                mode keeps the json objects where the center is outside the crop but parts of the
                objects are inside the crop. For these cases creates a new json object with a new center.
                Warning: mode "keep_objects" is still not included.
    """
    if crop_mode == "remove_objects":
        x, y = 0, 0
        if json_object["type"] == "polygon":
            contour = json2contours(json_object)
            x, y = get_center(contour)
        elif json_object["type"] == "ellipse":
            x, y = json_object["center"]["x"], json_object["center"]["y"]
        else:
            raise ValueError(f"No json type found of {json_object['type']}")

        within_crop = True
        if x < new_x[0] or x > new_x[1] or y < new_y[0] or y > new_y[1]:
            within_crop = False
    else:
        within_crop = True

    return within_crop


def crop_json(
    json_list: List[dict],
    x_translation: int,
    y_translation: int,
    crop_dim: Union[list, tuple],
    crop_mode: str = "remove_objects",
) -> List[dict]:
    """
    Takes a list of json objects (ellipse or polygon, standard cytosmart format)
    and removes or/and translates json object in the list. If the center of a
    json object lays outside the croped dimensions it removes the json objects.
    json_list: (list[dict]) a list of jsons in standard cytosmart format
    x_translation: (int) The translation in the x-direction of the cropped image
                relative to the original image
    y_translation: (int) The translation in the y-direction of the cropped image
                relative to the original image
    crop_dim: (list or tuple) (x,y) dimensions of the the cropped image
            that relate to the list of jsons
    crop_mode: (string) "remove_objects" & "keep_objects". Default mode is 'remove_objects'
                using this mode removes json_object if center is outside the crop. "keep_objects"
                mode keeps the json objects where the center is outside the crop but parts of the
                objects are inside the crop. For these cases creates a new json object with a new center.
                Warning: mode "keep_objects" is still not included.

    input_json_list = [
            {
                "type": "ellipse",
                "center": {"x": 2, "y": 3},
                "radiusX": 5,
                "radiusY": 5,
                "angleOfRotation": 0,
            },
            {   "type": "ellipse",
                "center": {"x": 4, "y": 7},
                "radiusX": 6,
                "radiusY": 6},
        ]

    crop_dim = (5, 6)
    x_translation = 2
    y_translation = 0

    expected_json_list = [
        {
            "type": "ellipse",
            "center": {"x": 0, "y": 3},
            "radiusX": 5,
            "radiusY": 5,
            "angleOfRotation": 0,
        }
    ]
    """

    new_json_list = []

    json_list_copy = copy.deepcopy(json_list)

    for c, json_object in enumerate(json_list):
        json_copy = copy.deepcopy(json_object)
        to_keep = json_object_to_keep(
            json_copy,
            (x_translation, x_translation + crop_dim[0]),
            (y_translation, y_translation + crop_dim[1]),
            crop_mode,
        )
        if to_keep:
            json_list_copy[c] = translation_json(
                json_copy, -x_translation, -y_translation
            )

        else:
            continue

        new_json_list.append(json_list_copy[c])

    return new_json_list



from dataclasses import asdict
from typing import List, Tuple, Union, Dict
import cv2
import numpy as np

from tomni.annotation_manager.annotations.annotation import Annotation
from tomni.annotation_manager.annotations.point import Point
from tomni.annotation_manager.utils import (
    are_lines_equal,
    parse_points_to_contour,
    overlap_object,
    parse_points_to_inner_contour,
)

MIN_NR_POINTS_POLYGON = 5

from ...utils import compress_polygon_points, parse_points_to_contour


class Polygon(Annotation):
    def __init__(
        self,
        points: List[Point],
        id: str,
        inner_points: Union[List[List[Point]], List] = [],
        label: str = "",
        children: List[Annotation] = [],
        parents: List[Annotation] = [],
        feature_multiplier: int = 1,
        accuracy: float = 1,
    ):
        """Initializes a Polygon object.

        Args:
            points (List[Point]): Collection of edges describing the polygon.
            id (str): UUID identifier.
            inner_points (Union[List[List[Point]], List]): Collection of edges describing the inner contours or an empty list when there are no inner contours.
            label (str): Class label of annotation.
            children (List[Annotation]): Tracking annotations. Refers to t+1.
            parents (List[Annotation]): Tracking annotations. Refers to t-1.
            accuracy (float, optional): The confidence of the model's prediction. Defaults to 1.
            features (Union[List[str],None]): list of features that the user wants returned.
                Defaults to None
            metric_unit (str, optional): A suffix added to the name of the feature in the dict. Defaults to "".
            feature_multiplier (int, optional): Pixel density of the image. Defaults to 1.
        """
        super().__init__(id, label, children, parents, accuracy)
        self._points = points
        self._inner_points = inner_points
        if len(self._points) < 5:
            raise ValueError("Polygon must have atleast 5 points.")
        self._contour: np.ndarray = parse_points_to_contour(points)
        self._inner_contours: List[np.ndarray] = parse_points_to_inner_contour(
            inner_points
        )
        self._feature_multiplier = feature_multiplier

        self._area: Union[float, None] = None
        self._outer_area: Union[float, None] = None
        self._aspect_ratio: Union[float, None] = None
        self._average_diameter: Union[float, None] = None
        self._circularity: Union[float, None] = None
        self._convex_hull_area: Union[float, None] = None
        self._major_axis: Union[float, None] = None
        self._minor_axis: Union[float, None] = None
        self._perimeter: Union[float, None] = None
        self._roundness: Union[float, None] = None

        self._all_features = {
            "area": {"is_ratio": False},
            "aspect_ratio": {"is_ratio": True},
            "average_diameter": {"is_ratio": False},
            "circularity": {"is_ratio": True},
            "convex_hull_area": {"is_ratio": False},
            "major_axis": {"is_ratio": False},
            "minor_axis": {"is_ratio": False},
            "perimeter": {"is_ratio": False},
            "roundness": {"is_ratio": True},
        }

    @property
    def accuracy(self) -> float:
        """Accuracy of ellipse."""
        return self._accuracy

    @accuracy.setter
    def accuracy(self, *arg, **kwargs) -> None:
        raise SyntaxError("Accuracy is Immutable")

    @property
    def label(self):
        return super().label

    @label.setter
    def label(self, value) -> None:
        super().label = value

    @property
    def points(self) -> List[Point]:
        return self._points

    @points.setter
    def points(self, *arg, **kwargs) -> None:
        raise SyntaxError("Points are Immutable")

    @property
    def inner_points(self) -> List[List[Point]]:
        return self._inner_points

    @inner_points.setter
    def inner_points(self, *arg, **kwargs) -> None:
        raise SyntaxError("Inner Points are Immutable")

    @property
    def area(self) -> Union[float, None]:
        """Quantity that expresses the extent of a polygon determined by cv2 contour operations.
        Requires 5 or more points.

        Returns:
            Union[float, None]: Polygon's area or None.
        """
        if self._area is None:
            self._calculate_area()
        return self._area * self._feature_multiplier**2

    @property
    def circularity(self) -> Union[float, None]:
        """Circularity: (4 * pi * Area) / perimeter ** 2)

        Returns:
            Union[float, None]: Circularity in [0, 1] or None.
        """

        if self._circularity is None:
            self._calculate_circularity()
        return self._circularity

    @property
    def convex_hull_area(self) -> Union[float, None]:
        """Convex Hull Area by cv2 contour operations.

        Returns:
            Union[float, None]: Polygon's convex hull area or None.
        """

        if self._convex_hull_area is None:
            self._calculate_convex_hull_area()
        return self._convex_hull_area * self._feature_multiplier**2

    @property
    def average_diameter(self) -> float:
        """Returns the average diameter of a polygon.

        Returns:
            float: Average diameter.
        """
        if self._average_diameter is None:
            self._calculate_average_diameter()
        return self._average_diameter * self._feature_multiplier

    @property
    def minor_axis(self) -> float:
        """Return the minor axis of the polygon.

        Returns:
            float: Minor axis length.
        """
        if self._minor_axis is None:
            self._calculate_axes()
        return self._minor_axis * self._feature_multiplier

    @property
    def major_axis(self) -> float:
        """Return the major axis of the polygon.

        Returns:
            float: major axis length.
        """
        if self._major_axis is None:
            self._calculate_axes()
        return self._major_axis * self._feature_multiplier

    @property
    def aspect_ratio(self) -> float:
        """Ratio between minor and major axis.

        Returns:
            float: Ellipse's aspect ratio.
        """
        if self._aspect_ratio is None:
            self._calculate_aspect_ratio()
        return self._aspect_ratio

    @property
    def perimeter(self) -> Union[float, None]:
        """Total length of polygon's boundary determined by cv2 contour operations.

        Returns:
            Union[float, None]: Polygon's perimeter or None.
        """

        if self._perimeter is None:
            self._calculate_perimeter()
        return self._perimeter * self._feature_multiplier

    @property
    def roundness(self) -> Union[float, None]:
        """Roundness: Area / (radius_enclosing_circle**2 * pi).

        Returns:
            Union[float, None]: Polygon's roundness in [0, 1] or None.
        """

        if self._roundness is None:
            self._calculate_roundness()
        return self._roundness

    def to_dict(
        self,
        decimals: int = 2,
        features: Union[List[str], None] = None,
        metric_unit="",
        feature_multiplier: int = 1,
        **kwargs,
    ) -> dict:
        """Returns a dictionary of the annotation in cytosmart format.

        Args:
            decimals (int, optional): The number of decimals to use when rounding. Defaults to 2.
            features (Union[List[str], None], optional): The features that are calculated and returned in the dict.
                Defaults to None, which means all features are calculated and returned.
            metric_unit (str, optional): The suffic added to the dict key names in camelCasing. Defaults to "".
                For example: "area" with suffix "um" becomes "areaUm"
            feature_multiplier (int, optional): A multiplier used during feature calculations. For example; 1/742. Defaults to 1.

        Raises:
            ValueError: It raises and error if any of the features are not compatible with the Annotation Manager.

        Returns:
            dict: a dictionary of the annotation in cytosmart format with the calculated features.
        """
        self._feature_multiplier = feature_multiplier

        # Check if features list is provided; otherwise, return all features
        if features is None:
            features = list(self._all_features.keys())
        else:
            missing_features = set(features).difference(set(self._all_features.keys()))
            if missing_features:
                raise ValueError(
                    f"The following features are not compatible with the Annotation Manager: {', '.join(missing_features)}"
                )

        points = self._points.copy()
        inner_points = self._inner_points.copy()
        if kwargs.get("do_compress", False):
            points = compress_polygon_points(
                points,
                kwargs.get("epsilon", 3),
                min_number_of_points=MIN_NR_POINTS_POLYGON,
            )
            inner_points = [
                compress_polygon_points(
                    inner_polygon,
                    kwargs.get("epsilon", 3),
                    min_number_of_points=MIN_NR_POINTS_POLYGON,
                )
                for inner_polygon in inner_points
            ]

        inner_polygons = []
        for inner_polygon in inner_points:
            inner_polygons.append(
                [asdict(inner_point) for inner_point in inner_polygon]
            )

        polygon_dict = {
            "type": "polygon",
            "points": [asdict(point) for point in points],
            "inner_points": inner_polygons,
        }

        polygon_features = {}
        for feature in features:
            feature_name = (
                feature
                if self._all_features[feature]["is_ratio"]
                else feature + "_" + metric_unit
            )

            # Convert snake_casing to camelCasing
            first_word, *remaining_words = feature_name.split("_")
            feature_name = "".join(
                [first_word.lower(), *map(str.title, remaining_words)]
            )

            polygon_features[feature_name] = round(getattr(self, feature), decimals)
        polygon_dict = {**polygon_features, **polygon_dict}

        super_dict = super().to_dict(decimals=decimals)
        dict_return_value = {**super_dict, **polygon_dict}
        return dict_return_value

    def is_in_mask(self, mask_json: List[dict], min_overlap: float = 0.9):
        """Check if a polygon is within a binary mask.

        Args:
            mask_json (List[dict]): A list of dict masks in cytosmart dict format.
            min_overlap (float, optional): Minimum overlap required between the polygon a mask, expressed as a value between 0 and 1.
            Defaults to 0.9.

        Returns:
            bool: True if the polygon is within a mask and meets the required overlap, False otherwise.
        """
        if len(self.points) < 1:
            return False

        json_points = [{"x": point.x, "y": point.y} for point in self.points]
        json_object = {"type": "polygon", "points": json_points}

        for mask in mask_json:
            overlap_ratio = overlap_object(json_object, mask)
            if overlap_ratio >= min_overlap:
                return True

        return False

    def to_binary_mask(self, shape: Tuple[int, int]) -> np.ndarray:
        """Transform a polygon to a binary mask.

        Args:
            shape (Tuple[int, int]): Shape of the new polygon's binary mask.

        Returns:
            np.ndarray: A binary mask in [0, 1].
        """
        mask = np.zeros(shape, dtype=np.uint8)
        if len(self._points) > 0:
            points = np.array(
                [[point.x, point.y] for point in self._points], dtype=np.int32
            )
            cv2.fillPoly(mask, [points], color=1)

        if self._inner_points:
            mask_inner = np.zeros(shape, dtype=np.uint8)
            for inner_polygon in self._inner_points:
                points = np.array(
                    [[point.x, point.y] for point in inner_polygon], dtype=np.int32
                )
                cv2.fillPoly(mask_inner, [points], color=1)
            mask = mask - mask_inner

        return mask

    def _calculate_area(self) -> None:
        # outer area for calculating circularity f.e.
        self._outer_area = cv2.contourArea(self._contour)
        self._area = self._outer_area
        for inner_contour in self._inner_contours:
            area_inner = cv2.contourArea(inner_contour)
            self._area -= area_inner

    def _calculate_circularity(self):
        if not self._outer_area:
            self._calculate_area()

        if not self._perimeter:
            self._calculate_perimeter()

        self._circularity = (4 * np.pi * self._outer_area) / (self._perimeter**2)

    def _calculate_convex_hull_area(self) -> None:
        convex_hull = cv2.convexHull(self._contour)
        self._convex_hull_area = cv2.contourArea(convex_hull)

    def _calculate_perimeter(self) -> None:
        self._perimeter = cv2.arcLength(self._contour, True)

    def _calculate_roundness(self) -> None:
        if not self._outer_area:
            self._calculate_area()

        _, radius = cv2.minEnclosingCircle(self._contour)
        enclosing_circle_area = radius**2 * np.pi
        self._roundness = self._outer_area / enclosing_circle_area

    def _calculate_axes(self) -> None:
        _, (diameter_1, diameter_2), _ = cv2.fitEllipse(self._contour)
        self._minor_axis = min(diameter_1, diameter_2)
        self._major_axis = max(diameter_1, diameter_2)

    def _calculate_average_diameter(self) -> None:
        if self._minor_axis is None or self._major_axis is None:
            self._calculate_axes()

        self._average_diameter = (self._major_axis + self._minor_axis) / 2

    def _calculate_aspect_ratio(self) -> None:
        if self._minor_axis is None or self._major_axis is None:
            self._calculate_axes()

        self._aspect_ratio = self._minor_axis / self._major_axis

    def __eq__(self, other):
        if not isinstance(other, Polygon):
            # don't attempt to compare against unrelated types
            return False

        are_points_equal = are_lines_equal(self.points, other.points, is_enclosed=True)
        reverse_points = other.points
        reverse_points.reverse()
        are_points_equal_mirrored = are_lines_equal(
            self.points, reverse_points, is_enclosed=True
        )
        return are_points_equal | are_points_equal_mirrored



from ...shape_fitting.fit_rect_ellipse import fit_rect_around_ellipse


def json2bbox(scf_object: dict) -> tuple:
    """
    Converts the scf_object (polygon, ellipse or circle) into a bounding box coordinates (x_min, y_min, x_max, y_max)
    All bounding box values get round using python internal 'round' function (https://wiki.c2.com/?BankersRounding)
    E.g:
    scf_object = {
        "type": "polygon"
        "points" : [{'x' : 0, 'y' : 0}, {'x' : 20, 'y' : 0}, {'x' : 20, 'y' : 30},{'x' : 0, 'y' : 30}]
    }

    bbox = scf_object2bbox(scf_object) #(0, 0 , 20, 30)

    NB: The angle of rotation for an ellipse is taken into account

    Args:
        scf_object (dict): single standard CytoSMART format annotations. allowed types polygon, ellipse or circle

    Returns
        bounding box coordinates (tuple): (x_min, y_min, x_max, y_max)
    """

    if "type" in scf_object:
        if scf_object["type"] == "polygon":
            x_values, y_values = [val["x"] for val in scf_object["points"]], [
                val["y"] for val in scf_object["points"]
            ]
            return (
                int(round(min(x_values))),
                int(round(min(y_values))),
                int(round(max(x_values))),
                int(round(max(y_values))),
            )
        elif scf_object["type"] in ["circle", "ellipse"]:
            x1, y1, x2, y2 = fit_rect_around_ellipse(
                scf_object["center"]["x"],
                scf_object["center"]["y"],
                scf_object["radiusX"],
                scf_object["radiusY"],
                scf_object["angleOfRotation"],
            )
            return x1, y1, x2, y2
        else:
            raise ValueError(f"{scf_object['type']} is not a supported type")

    else:
        raise KeyError(
            f"The scf_object does not have the key 'type'. dict is in the wrong format. {scf_object}"
        )



import cv2
import numpy as np
from scipy.stats import kurtosis, skew


def rescale_intensity(img, min_val, max_val):
    output_img = np.clip(img, min_val, max_val)
    output_img = (output_img - min_val) / (max_val - min_val) * 255
    return output_img.astype(np.uint8)


def fluo_tophat(img, normalize=0):
    """
    This is the background subtraction of the fluorescent channel.
    This subtraction can be used for the green and the red channel
    It uses a morphological white tophat filter to create a map of the subtraction.

    input:
        img: A standard grayscale image
        normalize: Normalizes the image from 0 to 255, this looks better but data will be lost

    output:
        img: The result image after background subtraction
        p2 : The lower bound value where an x-amount of pixels are included
        p98: The higher bound value where an y-amount of pixels are included

    The p2 and p98 values are used to normalize the image, but won't be automatically applied
    due to loss of data. These values should be given later to the integration part so they can
    pass these values to the front-end.

    """
    img = img.astype(np.uint8)
    img = cv2.GaussianBlur(img, (5, 5), 0)
    filterSize = (25, 25)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, filterSize)
    img = cv2.morphologyEx(img, cv2.MORPH_TOPHAT, kernel, iterations=10)

    kur = np.mean(kurtosis(img, fisher=True))
    skew1 = np.mean(skew(img))
    if kur > 1 and skew1 > 1:
        p2, p98 = np.percentile(img, (15, 99.5), interpolation="linear")
        if normalize == 1:
            img = rescale_intensity(img, p2, p98)
    else:
        p2, p98 = np.percentile(img, (15, 100), interpolation="linear")
        if normalize == 1:
            img = rescale_intensity(img, p2, p98)

    return img, p2, p98



import numpy as np

from typing import List, Union, Tuple
from tomni.annotation_manager.annotations import Polygon, Point
import uuid
import cv2

MIN_NR_POINTS_POLYGON = 5
MIN_AREA_RETANGLE_POLYGON = 500


def _is_approx_rectangle(contour):
    contour_area = cv2.contourArea(contour)
    x, y, w, h = cv2.boundingRect(contour)
    bounding_rect_area = w * h

    if contour_area > 0 and bounding_rect_area > 0:
        similarity_ratio = contour_area / bounding_rect_area
        return similarity_ratio >= 0.75
    return False


def _add_point(contour):
    # Calculate the midpoint between the first and second points
    new_point = (contour[0][0] + contour[1][0]) / 2

    # Insert the new point between the first and second points
    new_contour = np.insert(contour, 1, new_point, axis=0)

    return new_contour


def contours2polygons(
    contours: Tuple[np.ndarray],
    include_inner_contours: bool = False,
    hierarchy: Union[np.ndarray, None] = None,
    label: str = "",
) -> List[Polygon]:
    """transforms contours from cv2 contours to Polygon objects.
    Contours' shape must be [N, 1, 2] with dtype of np.int32.

    Args:
        contours (Tuple(np.ndarray)): Collection of cv2 contours.
        include_inner_contours (bool, optional): whether to return inner contours.
        hierarchy (bool, optional): the hierarchy from cv2.findContours using the RETR_CCOMP mode.
            Defaults to None.
            !!!!!Currently, only the hierarchy returned by mode: RETR_CCOMP is supported!!!!!
            If none, no hierarchy will be used.
        label(str, optional): The label of the polygon. Defaults to "".
    Returns:
        List[np.ndarray]: A list of Polygon objects.
    """
    annotations = []

    if len(contours) == 0:
        return annotations

    # Check whether inner contours are present
    if include_inner_contours:
        if hierarchy is None:
            raise ValueError(
                "hierarchy must be provided if include_inner_contours is True"
            )
        # Iterate over all contours and their hierarchies
        for idx, contour in enumerate(contours):
            current_hierarchy = hierarchy[0][idx]
            if current_hierarchy[-1] == -1:
                if len(contour) < MIN_NR_POINTS_POLYGON:
                    if (
                        len(contour) == 4
                        and cv2.contourArea(contour) > MIN_AREA_RETANGLE_POLYGON
                        and _is_approx_rectangle(contour)
                    ):
                        contour = _add_point(contour)
                    else:
                        continue

                # change shape from [N, 1, 2] to [N, 2]
                contour = np.vstack(contour)
                outer_points = [Point(x=int(pt[0]), y=int(pt[1])) for pt in contour]

                # Find the indices of the inner contours
                inner_indices = [i for i, h in enumerate(hierarchy[0]) if h[3] == idx]
                # Add the corresponding inner contours
                inner_contours = [contours[inner_idx] for inner_idx in inner_indices]
                inner_contours = [
                    np.vstack(inner_contour) for inner_contour in inner_contours
                ]
                list_of_inner_points = [
                    [
                        Point(x=int(pt[0]), y=int(pt[1]))
                        for pt in inner_contour.reshape(-1, 2)
                    ]
                    for inner_contour in inner_contours
                ]

                annotations.append(
                    Polygon(
                        label=label,
                        id=str(uuid.uuid4()),
                        children=[],
                        parents=[],
                        points=outer_points,
                        inner_points=list_of_inner_points,
                    ),
                )
    elif not include_inner_contours:
        for contour in contours:
            if len(contour) < MIN_NR_POINTS_POLYGON:
                if (
                    len(contour) == 4
                    and cv2.contourArea(contour) > MIN_AREA_RETANGLE_POLYGON
                    and _is_approx_rectangle(contour)
                ):
                    contour = _add_point(contour)
                else:
                    continue
            contour = np.vstack(contour)
            points = [Point(x=int(pt[0]), y=int(pt[1])) for pt in contour]

            annotations.append(
                Polygon(
                    label=label,
                    id=str(uuid.uuid4()),
                    children=[],
                    parents=[],
                    points=points,
                    inner_points=[],
                ),
            )
    return annotations



from __future__ import absolute_import
from unittest import TestCase
import numpy as np
from numpy import array, uint8, uint16, uint32, int64
from .main import labels2listsOfPoints


class TestSelect_labels2ListOfPoints(TestCase):
    def test_select_labels_as_indices(self):
        data = array(
            [
                [0, 5, 5, 0, 0, 0],
                [0, 0, 0, 0, 3, 3],
                [1, 0, 0, 0, 0, 0],
                [0, 2, 0, 0, 0, 0],
                [0, 2, 2, 0, 0, 0],
                [0, 2, 2, 0, 8, 0],
                [0, 2, 2, 0, 0, 0],
            ]
        )

        expected_output = [
            [[0, 2]],
            [[1, 3], [1, 4], [1, 5], [1, 6], [2, 4], [2, 5], [2, 6]],
            [[4, 1], [5, 1]],
            [[1, 0], [2, 0]],
            [[4, 5]],
        ]

        result = labels2listsOfPoints(data)

        self.assertEqual(len(expected_output), len(result))

        for i in range(len(expected_output)):
            x = result[i]
            print(x)
            np.testing.assert_array_equal(x, expected_output[i])



from unittest import TestCase
from .main import approximate_circle_by_area
import numpy as np


class TestGetAppoxCircleArea(TestCase):
    def test_free_shape(self):
        contour = np.array([[[70, 120]], [[80, 80]], [[70, 40]], [[60, 80]]])

        expected_result = (70, 80, 15.9576912161)
        result = approximate_circle_by_area(contour)
        print(result)

        self.assertEqual(len(result), len(expected_result))
        for i in range(len(result)):
            self.assertAlmostEqual(result[i], expected_result[i], 4)



import numpy as np


def json2contours(json: dict) -> np.ndarray:
    """
    Set a default json format object to a contour.
    json: (dict) the objects needs:
        'points': (if polygon) list of dict of form {x: ..., y: ...}
        'type': needs to be polygon
    """
    if json.get("type", "").lower() == "polygon":
        result = [[[i["x"], i["y"]]] for i in json["points"]]
    else:
        raise ValueError(
            f"The type {json.get('type', 'NO TYPE GIVEN')} is not supported"
        )
    return np.array(result, dtype=np.int32)



from unittest import TestCase

from tomni.annotation_manager.utils.contours2polygons.main import (
    _is_approx_rectangle,
    _add_point,
    contours2polygons,
)
from tomni.annotation_manager.annotations import Polygon, Point
import numpy as np
import uuid


class TestContours2Polygons(TestCase):
    def test_happy_flow(self) -> None:
        contours = (
            np.array(
                [[[0, 0]], [[10, 0]], [[100, 0]], [[100, 100]], [[50, 100]], [[0, 100]]]
            ),
        )
        polygons = contours2polygons(contours, None)

        expected = [
            Polygon(
                label="",
                id=str(uuid.uuid4()),
                children=[],
                parents=[],
                points=[
                    Point(0, 0),
                    Point(10, 0),
                    Point(100, 0),
                    Point(100, 100),
                    Point(50, 100),
                    Point(0, 100),
                ],
                inner_points=[],
            )
        ]

        self.assertEqual(polygons, expected)

    def test_happy_flow_label(self) -> None:
        contours = (
            np.array(
                [[[0, 0]], [[10, 0]], [[100, 0]], [[100, 100]], [[50, 100]], [[0, 100]]]
            ),
        )
        polygons = contours2polygons(contours, None, label="labelled")

        expected = [
            Polygon(
                label="labelled",
                id=str(uuid.uuid4()),
                children=[],
                parents=[],
                points=[
                    Point(0, 0),
                    Point(10, 0),
                    Point(100, 0),
                    Point(100, 100),
                    Point(50, 100),
                    Point(0, 100),
                ],
                inner_points=[],
            )
        ]

        self.assertEqual(polygons, expected)

    def test_happy_flow_hierarchy(self) -> None:
        contours = (
            np.array([[[1, 1]], [[1, 8]], [[7, 7]], [[7, 8]], [[7, 1]]]),
            np.array(
                [
                    [[2, 3]],
                    [[3, 2]],
                    [[5, 2]],
                    [[6, 3]],
                    [[6, 6]],
                    [[5, 7]],
                    [[3, 7]],
                    [[2, 6]],
                ],
            ),
        )

        hierarchy = np.array(
            [
                [
                    [-1, -1, 1, -1],
                    [-1, -1, -1, 0],
                ]
            ]
        )

        polygons = contours2polygons(
            contours=contours,
            hierarchy=hierarchy,
            include_inner_contours=True,
            label="labelled",
        )

        expected = [
            Polygon(
                label="labelled",
                id=str(uuid.uuid4()),
                children=[],
                parents=[],
                points=[
                    Point(1, 1),
                    Point(1, 8),
                    Point(7, 7),
                    Point(7, 8),
                    Point(7, 1),
                ],
                inner_points=[
                    [
                        Point(2, 3),
                        Point(3, 2),
                        Point(5, 2),
                        Point(6, 3),
                        Point(6, 6),
                        Point(5, 7),
                        Point(3, 7),
                        Point(2, 6),
                    ]
                ],
            ),
        ]

        self.assertEqual(polygons, expected)

    def test_happy_flow_multiple_contours(self) -> None:
        contours = (
            np.array(
                [[[0, 0]], [[10, 0]], [[100, 0]], [[100, 100]], [[50, 100]], [[0, 100]]]
            ),
            np.array(
                [[[0, 0]], [[10, 0]], [[200, 0]], [[200, 100]], [[50, 100]], [[0, 100]]]
            ),
        )

        polygons = contours2polygons(contours, None, label="labelled")

        expected = [
            Polygon(
                label="labelled",
                id=str(uuid.uuid4()),
                children=[],
                parents=[],
                points=[
                    Point(0, 0),
                    Point(10, 0),
                    Point(100, 0),
                    Point(100, 100),
                    Point(50, 100),
                    Point(0, 100),
                ],
                inner_points=[],
            ),
            Polygon(
                label="labelled",
                id=str(uuid.uuid4()),
                children=[],
                parents=[],
                points=[
                    Point(0, 0),
                    Point(10, 0),
                    Point(200, 0),
                    Point(200, 100),
                    Point(50, 100),
                    Point(0, 100),
                ],
                inner_points=[],
            ),
        ]

        self.assertEqual(polygons, expected)

    def test_include_inner_and_hierarchy(self) -> None:
        contours = (
            np.array(
                [[[0, 0]], [[10, 0]], [[100, 0]], [[100, 100]], [[50, 100]], [[0, 100]]]
            ),
        )
        hierarchy = np.array(
            [
                [
                    [-1, -1, 1, -1],
                    [-1, -1, -1, 0],
                ]
            ]
        )

        polygons = contours2polygons(
            contours=contours, include_inner_contours=False, hierarchy=hierarchy
        )

        expected = [
            Polygon(
                label="",
                id=str(uuid.uuid4()),
                children=[],
                parents=[],
                points=[
                    Point(0, 0),
                    Point(10, 0),
                    Point(100, 0),
                    Point(100, 100),
                    Point(50, 100),
                    Point(0, 100),
                ],
                inner_points=[],
            )
        ]

        self.assertEqual(polygons, expected)

    def test_external_hierarchy_include_inner_contours_true(self) -> None:
        contours = (
            np.array(
                [[[0, 0]], [[10, 0]], [[100, 0]], [[100, 100]], [[50, 100]], [[0, 100]]]
            ),
        )
        hierarchy = np.array([[[1, -1, -1, -1]]])

        polygons = contours2polygons(
            contours=contours, include_inner_contours=True, hierarchy=hierarchy
        )

        expected = [
            Polygon(
                label="",
                id=str(uuid.uuid4()),
                children=[],
                parents=[],
                points=[
                    Point(0, 0),
                    Point(10, 0),
                    Point(100, 0),
                    Point(100, 100),
                    Point(50, 100),
                    Point(0, 100),
                ],
                inner_points=[],
            )
        ]

        self.assertEqual(polygons, expected)

    def test_include_inner_and_hierarchy_mismatch(self) -> None:
        contours = (
            np.array(
                [[[0, 0]], [[10, 0]], [[100, 0]], [[100, 100]], [[50, 100]], [[0, 100]]]
            ),
        )

        self.assertRaises(
            ValueError,
            contours2polygons,
            contours,
            include_inner_contours=True,
            hierarchy=None,
        )

    def test_too_few_points(self) -> None:
        contours = np.array(
            [[[0, 0]], [[5, 0]], [[5, 5]], [[0, 5]]],
        )
        polygons = contours2polygons(contours, None)

        expected = []

        self.assertEqual(polygons, expected)

    def test_is_approx_rectangle_true(self) -> None:
        contours = np.array(
            [
                [[0, 0]],
                [[100, 0]],
                [[100, 100]],
                [[0, 100]],
            ],
        )
        polygons = _is_approx_rectangle(contours)

        expected = True

        self.assertEqual(polygons, expected)

    def test_is_approx_rectangle_false(self) -> None:
        contour = np.array(
            [
                [[0, 0]],
                [[50, 25]],
                [[100, 125]],
                [[50, 100]],
            ],
        )
        polygons = _is_approx_rectangle(contour)

        expected = False

        self.assertEqual(polygons, expected)

    def test_is_approx_rectangle_false_diamond(self) -> None:
        contour = np.array(
            [
                [[50, 0]],
                [[0, 50]],
                [[50, 100]],
                [[100, 50]],
            ],
        )
        polygons = _is_approx_rectangle(contour)

        expected = False

        self.assertEqual(polygons, expected)

    def test_add_point(self) -> None:
        contour = np.array(
            [
                [[0, 0]],
                [[50, 0]],
                [[50, 50]],
                [[0, 50]],
            ],
        )
        output = _add_point(contour)

        expected = np.array(
            [
                [[0, 0]],
                [[25, 0]],
                [[50, 0]],
                [[50, 50]],
                [[0, 50]],
            ],
        )

        np.testing.assert_array_equal(output, expected)

    def test_add_point_2(self) -> None:
        contour = np.array(
            [
                [[0, 0]],
                [[100, 0]],
                [[100, 100]],
                [[0, 100]],
            ],
        )
        output = _add_point(contour)

        expected = np.array(
            [
                [[0, 0]],
                [[50, 0]],
                [[100, 0]],
                [[100, 100]],
                [[0, 100]],
            ],
        )

        np.testing.assert_array_equal(output, expected)



import uuid
import numpy as np
from typing import List, Union


def contours2json(contours: Union[np.ndarray, List[list]]) -> List[dict]:
    """
    Transforms a list or array of contours into list of jsons
    contours: (numpy.array, list) needs to be the contours produced by opencv
    """
    result = []
    for contour in contours:
        shape = {"type": "polygon", "points": []}
        for point in contour:
            shape["points"].append(dict(x=int(point[0][0]), y=int(point[0][1])))
        shape["id"] = str(uuid.uuid4())
        shape["parents"] = []
        shape["children"] = []
        result.append(shape)
    return result



import cv2
import numpy as np


def roundness(contour: np.ndarray) -> float:
    r"""
    Returns roundess of the contour
    Roundness is determined by difference in areas of the contour and enclosing circle
    Roundness = :math:`\frac{A}{r_e^2 \pi}`
    With :math:`r_e` being the radius of the enclosing circle.

    Args:
        contour (np.ndarray): An opencv contour of a single object

    Returns:
        float: roundess number between 0 and 1
    """

    area = cv2.contourArea(contour)
    _, radius = cv2.minEnclosingCircle(contour)
    enclosing_circle_area = radius**2 * np.pi
    roundness = area / enclosing_circle_area
    return roundness



from unittest import TestCase
from .main import crop_json


class TestRandomCropBbox(TestCase):
    def test_ellipse_x_movement(self):
        input_json = [
            {
                "type": "ellipse",
                "center": {"x": 2, "y": 3},
                "radiusX": 5,
                "radiusY": 5,
                "angleOfRotation": 0,
            },
            {"type": "ellipse", "center": {"x": 4, "y": 7}, "radiusX": 6, "radiusY": 6},
        ]

        inputDim = (5, 6)
        x_move = 2
        y_move = 0

        expected_json = [
            {
                "type": "ellipse",
                "center": {"x": 0, "y": 3},
                "radiusX": 5,
                "radiusY": 5,
                "angleOfRotation": 0,
            }
        ]

        result_json = crop_json(input_json, x_move, y_move, inputDim)
        self.assertListEqual(result_json, expected_json)

    def test_ellipse_xy_movement(self):

        input_json = [
            {
                "type": "ellipse",
                "center": {"x": 2, "y": 3},
                "radiusX": 5,
                "radiusY": 5,
                "angleOfRotation": 0,
            },
            {"type": "ellipse", "center": {"x": 4, "y": 8}, "radiusX": 6, "radiusY": 6},
        ]

        inputDim = (5, 6)
        x_move = 2
        y_move = 1

        expected_json = [
            {
                "type": "ellipse",
                "center": {"x": 0, "y": 2},
                "radiusX": 5,
                "radiusY": 5,
                "angleOfRotation": 0,
            }
        ]

        result_json = crop_json(input_json, x_move, y_move, inputDim)

        self.assertListEqual(result_json, expected_json)

        expected_input_json = [
            {
                "type": "ellipse",
                "center": {"x": 2, "y": 3},
                "radiusX": 5,
                "radiusY": 5,
                "angleOfRotation": 0,
            },
            {"type": "ellipse", "center": {"x": 4, "y": 8}, "radiusX": 6, "radiusY": 6},
        ]

        self.assertListEqual(input_json, expected_input_json)

    def test_polygon_x_movement(self):

        # [p, bx, by, bw, bh, c1, c2, c3]
        input_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 2},
                ],
            },
            {
                "type": "polygon",
                "points": [
                    {"x": 4, "y": 6},
                    {"x": 6, "y": 6},
                    {"x": 6, "y": 4},
                    {"x": 4, "y": 6},
                ],
            },
        ]

        inputDim = (5, 6)
        x_move = 2
        y_move = 0

        expected_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 2, "y": 6},
                    {"x": 4, "y": 6},
                    {"x": 4, "y": 4},
                    {"x": 2, "y": 6},
                ],
            }
        ]

        expected_input_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 2},
                ],
            },
            {
                "type": "polygon",
                "points": [
                    {"x": 4, "y": 6},
                    {"x": 6, "y": 6},
                    {"x": 6, "y": 4},
                    {"x": 4, "y": 6},
                ],
            },
        ]
        result_json = crop_json(input_json, x_move, y_move, inputDim)
        self.assertListEqual(result_json, expected_json)

        self.assertEqual(len(input_json), len(expected_input_json))
        for i in range(len(input_json)):
            self.assertDictEqual(input_json[i], expected_input_json[i])

    def test_polygon_xy_movement(self):

        input_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 2},
                ],
            },
            {
                "type": "polygon",
                "points": [
                    {"x": 4, "y": 6},
                    {"x": 6, "y": 6},
                    {"x": 6, "y": 4},
                    {"x": 4, "y": 6},
                ],
            },
        ]

        inputDim = (5, 6)
        x_move = 2
        y_move = 1

        expected_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 2, "y": 5},
                    {"x": 4, "y": 5},
                    {"x": 4, "y": 3},
                    {"x": 2, "y": 5},
                ],
            }
        ]

        result_json = crop_json(input_json, x_move, y_move, inputDim)
        self.assertListEqual(result_json, expected_json)

    def test_expected_polygon_no_crop(self):

        input_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 2},
                ],
            },
            {
                "type": "polygon",
                "points": [
                    {"x": 4, "y": 6},
                    {"x": 6, "y": 6},
                    {"x": 6, "y": 4},
                    {"x": 4, "y": 6},
                ],
            },
        ]

        inputDim = (10, 10)
        x_move = 0
        y_move = 0

        expected_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 2},
                ],
            },
            {
                "type": "polygon",
                "points": [
                    {"x": 4, "y": 6},
                    {"x": 6, "y": 6},
                    {"x": 6, "y": 4},
                    {"x": 4, "y": 6},
                ],
            },
        ]

        result_json = crop_json(input_json, x_move, y_move, inputDim)
        self.assertListEqual(result_json, expected_json)

    def test_poly_GONE(self):
        input_json = [
            {
                "type": "polygon",
                "points": [
                    {"x": 0, "y": 2},
                    {"x": 2, "y": 2},
                    {"x": 2, "y": 0},
                    {"x": 0, "y": 2},
                ],
            },
            {
                "type": "polygon",
                "points": [
                    {"x": 4, "y": 6},
                    {"x": 6, "y": 6},
                    {"x": 6, "y": 4},
                    {"x": 4, "y": 6},
                ],
            },
        ]

        inputDim = (2, 2)
        x_move = 8
        y_move = 0

        expected_json = []

        result_json = crop_json(input_json, x_move, y_move, inputDim)
        self.assertListEqual(result_json, expected_json)



import numpy as np
import cv2
import warnings


def absolute_difference(
    img: np.ndarray, gauss_size: int = 21, new_median: int = 128
) -> np.ndarray:
    """
    Remove big blurry artifact from the image.
    It will do this by taking the absolute difference between the blurred and raw image.

    # Usage
    Best not to use this for human vision.
    The algorithm can have weird artifacts.
    This is good as algorithm input for algorithms that rely on absolute pixel difference.

    # Fractal
    This algorithm is fractal.
    Meaning the illumination correction followed by cropping will given the same result than first cropping.

    img (np.array): input image
    gauss_size (int): needs to be an odd number.
        This is corralated to the size of the big artifact to remove.
    """

    img = np.array(img, dtype=np.int16)

    if gauss_size % 2 == 0:
        gauss_size += 1  # Only works with odd gauss size
        warnings.warn(
            "the gauss_size needs to be odd. It is {}. One is added to it".format(
                gauss_size
            ),
            SyntaxWarning,
        )

    """
    The illumination is every detail bigger than the gauss_size.
    This includes the inside of square bigger than gauss_size, but not its borders.
    """
    img -= cv2.GaussianBlur(img, (gauss_size, gauss_size), 0)

    img += new_median

    # All overhead needs to be rounded down
    img = np.clip(img, 0, 255)
    return img.astype(np.uint8)



from unittest import TestCase
import numpy as np
from .main import roundness


class TestRoundness(TestCase):
    def test_square(self):
        contour = np.array([[[500, 50]], [[500, 100]], [[550, 100]], [[550, 50]]])

        expectedResult = (50 * 50) / (25 * 25 * 2 * np.pi)

        result = roundness(contour)

        self.assertAlmostEqual(result, expectedResult, 5)

    def test_octagon(self):
        # Octagon is rounder then a square
        # This is not perfect but okay
        contour = np.array(
            [
                [[500, 50]],
                [[490, 75]],
                [[500, 99]],
                [[525, 109]],
                [[549, 99]],
                [[559, 75]],
                [[549, 50]],
                [[525, 40]],
            ]
        )

        expectedResult = 0.8964594850343249

        result = roundness(contour)

        self.assertAlmostEqual(result, expectedResult, 5)

    def test_square_with_dent(self):
        """
        A dent should not impact roundness to much as long as the area of the dent is relatively small
        """
        contour = np.array(
            [
                [[500, 50]],
                [[500, 100]],
                # start dent
                [[524, 100]],
                [[524, 51]],
                [[525, 51]],
                [[525, 100]],
                # end ent
                [[550, 100]],
                [[550, 50]],
            ]
        )

        expectedResult = (50 * 50 - 49) / (25 * 25 * 2 * np.pi)

        result = roundness(contour)

        self.assertAlmostEqual(result, expectedResult, 5)

    def test_flat_rectangle(self):
        # A rectangle of 450 by 50
        contour = np.array([[[0, 50]], [[0, 100]], [[450, 100]], [[450, 50]]])

        expectedResult = 0.13974567201687108

        result = roundness(contour)

        self.assertAlmostEqual(result, expectedResult)



from unittest import TestCase
from .main import translation_json


class TestScaleObject(TestCase):
    def test_ellipses(self):
        inputDict = {
            "type": "ellipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 10,
            "angleOfRotation": 0,
            "id": "unicorn",
        }
        translation_x = 50
        translation_y = -100

        expectedDict = {
            "type": "ellipse",
            "center": {"x": 70, "y": -70},
            "radiusX": 10,
            "radiusY": 10,
            "angleOfRotation": 0,
            "id": "unicorn",
        }

        result = translation_json(inputDict, translation_x, translation_y)

        self.assertDictEqual(result, expectedDict)

    def test_polygon(self):
        inputDict = {
            "type": "polygon",
            "points": [
                {"x": 500, "y": 52},
                {"x": 500, "y": 100},
                {"x": 540, "y": 100},
                {"x": 540, "y": 52},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
        }
        translation_x = -150
        translation_y = 10

        expectedDict = {
            "type": "polygon",
            "points": [
                {"x": 350, "y": 62},
                {"x": 350, "y": 110},
                {"x": 390, "y": 110},
                {"x": 390, "y": 62},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
        }

        result = translation_json(inputDict, translation_x, translation_y)

        self.assertDictEqual(result, expectedDict)

    def test_unknown_type(self):
        inputDict = {
            "type": "unicornType",
        }
        translation_x = -150
        translation_y = 10
        with self.assertRaises(ValueError):
            result = translation_json(inputDict, translation_x, translation_y)



import uuid
from typing import Dict, List, Tuple, Union

import cv2
import numpy as np

from tomni.annotation_manager.utils.contours2polygons import contours2polygons
from .annotations import Annotation, Ellipse, Point, Polygon
from .utils import parse_points_to_contour

MIN_NR_POINTS_POLYGON = 5


class AnnotationManager(object):
    def __init__(self, annotations: List[Annotation]):
        """Initializes a AnnotationManager object.

        Args:
            annotations (List[Annotation]): Collection of annotations, e.g. polygon or ellipse.
        """
        self._annotations = annotations

    @classmethod
    def from_dicts(
        cls,
        dicts: List[dict],
    ):
        TYPE_KEY = "type"
        LABEL_KEY = "label"
        CHILDREN_KEY = "children"
        PARENTS_KEY = "parents"
        ID_KEY = "id"
        CENTER_KEY = "center"
        annotations = []

        for d in dicts:
            if d[TYPE_KEY] == "ellipse":
                annotation = Ellipse(
                    label=d.get(LABEL_KEY, None),
                    id=d.get(ID_KEY, str(uuid.uuid4())),
                    children=d.get(CHILDREN_KEY, []),
                    parents=d.get(PARENTS_KEY, []),
                    radius_x=d["radiusX"],
                    radius_y=d.get("radiusY", None),
                    center=Point(x=d[CENTER_KEY]["x"], y=d[CENTER_KEY]["y"]),
                    rotation=d["angleOfRotation"],
                    accuracy=d.get("accuracy", 1),
                )
            elif d[TYPE_KEY] == "polygon":
                if len(d["points"]) < MIN_NR_POINTS_POLYGON:
                    continue
                if "inner_points" in d:
                    inner_points = [
                        [Point(x=pi["x"], y=pi["y"]) for pi in inner_contour]
                        for inner_contour in d["inner_points"]
                    ]
                else:
                    inner_points = []

                annotation = Polygon(
                    label=d.get(LABEL_KEY, None),
                    id=d.get(ID_KEY, str(uuid.uuid4())),
                    children=d.get(CHILDREN_KEY, []),
                    parents=d.get(PARENTS_KEY, []),
                    points=[Point(x=p["x"], y=p["y"]) for p in d["points"]],
                    inner_points=inner_points,
                    accuracy=d.get("accuracy", 1),
                )
            else:
                raise ValueError(
                    f"CDF cannot be created. Dict with id {d.get('id', None)} misses type-key with value ellipse or polygon."
                )
            annotations.append(annotation)

        return cls(annotations)

    @classmethod
    def from_binary_mask(
        cls, mask: np.ndarray, include_inner_contours: bool = False, label: str = ""
    ):
        """Initializes a AnnotationManager object from a binary mask.

        Args:
            mask (np.ndarray): Binary mask input.
            include_inner_contours (bool, optional): Include annotations that are contained within another annotation.
                Defaults to False.
        Returns:
            AnnotationManager: New annotation manager object from binary mask.
        """

        mask = mask.astype(np.uint8)
        mode = cv2.RETR_CCOMP if include_inner_contours else cv2.RETR_EXTERNAL
        contours, hierarchy = cv2.findContours(mask, mode, cv2.CHAIN_APPROX_SIMPLE)

        annotations = contours2polygons(
            contours=contours,
            hierarchy=hierarchy,
            include_inner_contours=include_inner_contours,
            label=label,
        )

        return cls(annotations)

    @classmethod
    def from_labeled_mask(
        cls,
        mask: np.ndarray,
        labels: Union[List[str], str] = "",
        include_inner_contours: bool = False,
    ):
        """Initializes a AnnotationManager object from a labeled mask.
        A labeled mask contains components indicated by the same pixel values (see example below).

        Example containing two components yielding a AnnotationManager object with two annotations:
        [[0,0,2,1,1],
        [0,0,2,0,0],
        [0,0,2,0,]]

        Args:
            mask (np.ndarray): A labeled mask with a max. nr. of components limited by max(np.uint32).
            classes(List[str], optional): A list of class names to add to Polygon labels. Defaults to None.
                Should have the same number of unique pixel values as classes.
                Class names in order of low pixel value to high pixel value.
            include_inner_contours (bool, optional): Include annotations that are contained within another annotation.
                Defaults to False.
        Returns:
            AnnotationManager: A new AnnotationManager object.
        """
        mode = cv2.RETR_CCOMP if include_inner_contours else cv2.RETR_EXTERNAL

        annotations = []
        if isinstance(labels, str):
            mask = mask.astype(np.uint8)
            _, mask = cv2.threshold(mask, 0, 255, cv2.THRESH_BINARY)
            contours, hierarchy = cv2.findContours(mask, mode, cv2.CHAIN_APPROX_SIMPLE)

            annotations = contours2polygons(
                contours=contours,
                include_inner_contours=include_inner_contours,
                hierarchy=hierarchy,
                label=labels,
            )

        elif isinstance(labels, List):
            unique_values = np.unique(mask)
            unique_values = unique_values[unique_values != 0]

            if len(labels) < len(unique_values):
                raise ValueError(
                    f"Not enough labels for unique pixel values. {len(labels)} labels for {len(unique_values)} unique pixel values."
                )

            # Generate seperate mask for each class and find contours.
            for pixel_value in unique_values:
                class_mask = np.uint8(mask == pixel_value)
                contours, hierarchy = cv2.findContours(
                    class_mask, mode, cv2.CHAIN_APPROX_SIMPLE
                )
                annotations.extend(
                    contours2polygons(
                        contours=contours,
                        hierarchy=hierarchy,
                        include_inner_contours=include_inner_contours,
                        label=labels[pixel_value - 1],
                    )
                )
        else:
            raise ValueError("Labels must be either a string or a list of strings.")

        return cls(annotations)

    @classmethod
    def from_darwin(cls, dicts: List[dict]):
        """must be an option"""
        pass

    @property
    def annotations(self) -> List[Annotation]:
        return self._annotations

    @annotations.setter
    def annotations(self, other_annotations: List[Annotation]):
        """I doubt this setter should be allowed to exist."""

    def __len__(self) -> int:
        return len(self._annotations)

    def __eq__(self, other: object) -> bool:
        """To check equality of to CDF objects
        Thats bit of a tricky one because you must compare all annotations IMO.
        Ex: cdf1 = AnnotationManager.from_something()
        cdf2 = AnnotationManager.from_something()
        is cdf1 == cdf2 must be possible.
        """
        pass

    def __contains__(self, other: Annotation):
        """to check if self.annotations contains other."""
        pass

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx < self.__len__():
            annotation = self.annotations[self.idx]
            self.idx += 1
            return annotation
        else:
            raise StopIteration

    def to_dict(
        self,
        decimals: int = 2,
        mask_json: Union[List[dict], None] = None,
        min_overlap: float = 0.9,
        features: Union[List[str], None] = None,
        metric_unit: str = "",
        feature_multiplier: int = 1,
        **kwargs,
    ) -> List[Dict]:
        """Transform AM object to a collection of our format.

        Args:
            decimals (int, optional): The number of decimals to use when rounding. Defaults to 2.
            mask_json (Union[dict, None], optional): The dict mask that indicates what area to include in the output dict.
                Defaults to None.
            min_overlap (float, optional): Minimum overlap required between the polygon and the mask, expressed as a value between 0 and 1
                Defaults to 0.9.
            features (Union[List[str], None], optional): The features that you want to calculate and add to the dict objects.
                Defaults to None, which returns all features.
            metric_unit (str, optional): The suffix you want to add to the dict keys' names in camelCasing. Defaults to "".
            feature_multiplier (int, optional): A multiplier used during feature calculation. For example 1/742. Defaults to 1.

        Returns:
            List[Dict]: Output is a list of dicts in cytosmart format.
        """

        if mask_json is not None:
            filtered_annotations = self._annotations.copy()
            filtered_annotations = [
                annotation
                for annotation in filtered_annotations
                if annotation.is_in_mask(mask_json, min_overlap)
            ]

            return [
                annotation.to_dict(
                    decimals=decimals,
                    features=features,
                    metric_unit=metric_unit,
                    feature_multiplier=feature_multiplier,
                    **kwargs,
                )
                for annotation in filtered_annotations
            ]

        return [
            annotation.to_dict(
                decimals=decimals,
                features=features,
                metric_unit=metric_unit,
                feature_multiplier=feature_multiplier,
                **kwargs,
            )
            for annotation in self._annotations
        ]

    def to_contours(self) -> List[np.ndarray]:
        """Transform AM object to a collection of cv2 contours.

        Raises:
            ValueError: Raises error when annotations are not of type `Polygon`.

        Returns:
            List[np.ndarray]: Collection of contours as [[[x_0, y_0],..., [x_n, y_n]], ... ,[[x_0, y_0],..., [x_m, y_m]]]
        """
        if not all(
            [isinstance(annotation, Polygon) for annotation in self._annotations]
        ):
            raise ValueError("`to_contours is only supported on polygon-annotations.`")

        contours = [
            parse_points_to_contour(annotation.points)
            for annotation in self._annotations
        ]

        return contours

    def to_binary_mask(self, shape: Tuple[int, int]) -> np.ndarray:
        """Transform an AM object to a binary mask.
        Annotations can only be polygon or ellipse.

        Args:
            shape (Tuple[int, int]): Shape of the new binary mask.

        Returns:
            np.ndarray: A binary mask in [0, 1].
        """
        mask = np.zeros(shape, dtype=np.uint8)
        for annotation in self.annotations:
            mask = cv2.bitwise_or(mask, annotation.to_binary_mask(shape))

        return mask

    def to_labeled_mask(self, shape: Tuple[int, int]) -> np.ndarray:
        """Transform an AM object to a labeled mask.
        Annotations can only be polygon or ellipse.

        shape (Tuple[int, int]): Shape of the new labeled mask.

        Returns:
            np.ndarray: A new labeled mask.
        """
        mask = np.zeros(shape, dtype=np.uint8)
        label_color = 1

        for annotation in self._annotations:
            if isinstance(annotation, Polygon):
                points = np.array(
                    [[point.x, point.y] for point in annotation.points], dtype=np.int32
                )
                cv2.fillPoly(mask, [points], color=label_color)
            elif isinstance(annotation, Ellipse):
                cv2.ellipse(
                    mask,
                    center=(annotation.center.x, annotation.center.y),
                    axes=(annotation.radius_x, annotation.radius_y),
                    angle=annotation.rotation,
                    startAngle=0,
                    endAngle=360,
                    color=label_color,
                    thickness=-1,
                )
            else:
                raise TypeError(
                    "Innapropiate annotation type for `to_labeled_mask`. Supported annotations are ellipse and polygon."
                )

            # increase color for every annotation.
            label_color += 1

        return mask

    def to_darwin(self) -> List[Dict]:
        """
        TODO: Convert annotations to darwin format (v7).
        """

    def __add__(self, other):
        """Ability to add to CDF objects together
        cdf1 + cdf2.

        or possiby
        - cdf + dict
        - cdf + darwin
        - ...
        """
        pass

    def __radd__(self, other):
        """Reverse of __add__
        if you do:
        dict + cdf -> error
        radd should flip the two parts and call add.
        so, dict + cdf becomes cdf + dict.
        """
        pass

    def delete_annotation(self, item: Annotation):
        """Remove an annotation from self.annotations.
        Find and delete.
        """
        pass

    def filter(
        self,
        feature: str,
        min_val: float,
        max_val: float,
        feature_multiplier: int = 1,
        inplace: bool = False,
    ):
        """Filter annotations by feature.

        Args:
            feature (str): Feature name, i.e. `roundness` or `area`.
            min_val (float): Minimum value to threshold.
            max_val (float): Maximum value to threshold
            feature_multiplier (int, optional): A multiplier used in the feature calculations. For example 1/742. Defaults to 1.
            inplace (bool, optional): If True, filter in-place. Modifies the object internally. If False, return collection of annotations. Defaults to False.

        Returns:
            AnnotationManager or List[Annotation]: Collection of filtered annotions or if `inplace=True` object with filterd annotations.
        """
        filtered_annotations = []

        for annotation in self._annotations:
            annotation._feature_multiplier = feature_multiplier
            feature_value = getattr(annotation, feature)
            if min_val <= feature_value <= max_val:
                filtered_annotations.append(annotation)

        if inplace:
            self._annotations = filtered_annotations
            return self

        return filtered_annotations

    @classmethod
    def get_circularity_summary(self):
        """loops the cdf items to get avg, std, min, max."""

    def get_feature_summaries(self, features: List[str]) -> Dict:
        """Pass a list of features to calculated.

        This is not ideal and i am for suggestion. I want to avoid having to calculate over and over.
        """
        circularity = []
        for cdf_item in self._cdf_data:
            circularity.append(cdf_item.circularity)

        return {
            "circularity": {"avg": 1, "std": 1, "min": 1, "max": 1},
            "...": "...",
            "feature_n": {"avg_n": 1, "std_n": 1, "min_n": 1, "max_n": 1},
        }



from unittest import TestCase
import numpy as np
from .main import json2bbox
from ...shape_fitting.fit_rect_ellipse import fit_rect_around_ellipse


class TestJson2bbox(TestCase):
    def test_happy_flow_polygon(self):
        expectedResult = (500, 50, 539, 99)

        inputContours = {
            "type": "polygon",
            "points": [
                {"x": 500, "y": 50},
                {"x": 500, "y": 99},
                {"x": 520, "y": 79},
                {"x": 539, "y": 99},
                {"x": 520, "y": 79},
                {"x": 539, "y": 50},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
        }

        result = json2bbox(inputContours)

        np.testing.assert_equal(result, expectedResult)

    def test_happy_flow_ellipse(self):
        expectedResult = (0, 1, 12, 9)

        inputContours = {
            "type": "ellipse",
            "center": {"x": 6.0, "y": 5.0},
            "radiusX": 6.0,
            "radiusY": 4.0,
            "angleOfRotation": 0,
            "id": "unicorn",
        }

        result = fit_rect_around_ellipse(
            inputContours["center"]["x"],
            inputContours["center"]["y"],
            inputContours["radiusX"],
            inputContours["radiusY"],
            inputContours["angleOfRotation"],
        )

        np.testing.assert_equal(result, expectedResult)

    def test_happy_flow_ellipse_with_angle(self):
        expectedResult = (0, 0, 12, 10)

        inputContours = {
            "type": "ellipse",
            "center": {"x": 6.0, "y": 5.0},
            "radiusX": 6.0,
            "radiusY": 4.0,
            "angleOfRotation": 30,
            "id": "unicorn",
        }

        result = fit_rect_around_ellipse(
            inputContours["center"]["x"],
            inputContours["center"]["y"],
            inputContours["radiusX"],
            inputContours["radiusY"],
            inputContours["angleOfRotation"],
        )
        np.testing.assert_almost_equal(result, expectedResult)

    def test_happy_flow_ellipse_90(self):
        expectedResult = (2, -1, 10, 11)

        inputContours = {
            "type": "ellipse",
            "center": {"x": 6.0, "y": 5.0},
            "radiusX": 6.0,
            "radiusY": 4.0,
            "angleOfRotation": 90,
            "id": "unicorn",
        }

        result = fit_rect_around_ellipse(
            inputContours["center"]["x"],
            inputContours["center"]["y"],
            inputContours["radiusX"],
            inputContours["radiusY"],
            inputContours["angleOfRotation"],
        )
        np.testing.assert_almost_equal(result, expectedResult)

    def test_happy_flow_ellipse_180(self):
        expectedResult = (0, 1, 12, 9)

        inputContours = {
            "type": "ellipse",
            "center": {"x": 6.0, "y": 5.0},
            "radiusX": 6.0,
            "radiusY": 4.0,
            "angleOfRotation": 180,
            "id": "unicorn",
        }

        result = fit_rect_around_ellipse(
            inputContours["center"]["x"],
            inputContours["center"]["y"],
            inputContours["radiusX"],
            inputContours["radiusY"],
            inputContours["angleOfRotation"],
        )

        np.testing.assert_almost_equal(result, expectedResult)

    def test_value_error_unsupported_type(self):
        inputContours = {
            "type": "unicorn",
            "points": [
                {"x": 500, "y": 50},
                {"x": 500, "y": 99},
                {"x": 539, "y": 99},
                {"x": 539, "y": 50},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
        }

        with self.assertRaises(ValueError):
            json2bbox(inputContours)

    def test_value_error_no_type_given(self):
        inputContours = {
            "points": [
                {"x": 500, "y": 50},
                {"x": 500, "y": 99},
                {"x": 539, "y": 99},
                {"x": 539, "y": 50},
            ],
            "id": "unicorn",
            "parents": [],
            "children": [],
        }

        with self.assertRaises(KeyError):
            json2bbox(inputContours)



import numpy as np


def img_dim(arr, showChannels=False):
    """
    This function gives the image dimensions and optional channel of a numpy array.
    arr: (numpy.ndarray) the image represented as numpy array
    showChannels: (boolean) if you want to output the number of color channels e.q. 1 for greyscale, 3 for BGR, 4 for BGRA
    """
    s = np.shape(arr)

    if showChannels:
        if len(s) > 2:
            return s[1], s[0], s[2]
        else:
            return s[1], s[0], 1
    else:
        return s[1], s[0]



from typing import Union
import numpy as np
import cv2


def convert_color(img: np.ndarray, new_color: Union[str, int]) -> np.ndarray:
    """
    Convert a the color of an image to a wanted color, without needed to know the current color.
    It will assume only gray, BGR, and BGRA exists (these are the common openCV color types)

    Args:
        img (np.ndarray uint8): The image that needs to be converted. This image is gray, BGR or BGRA
        new_color (Union[str, int]): Color the new image should be gray (1), BGR (2) or BGRA (3)

    Raises:
        ValueError: new_color is not supported
        ValueError: img has more then 3 dimensions or less then 2
        ValueError: img as a number of channels that is not 1, 3, or 4

    Returns:
        np.ndarray: converted image
    """
    img = img.astype(np.uint8)
    # Not getting into that debate
    if type(new_color) is str:
        if new_color.upper() in ["GRAY", "GREY"]:
            new_color = 1
        # Also not getting into that debate
        elif new_color.upper() in ["BGR", "COLOR", "COLOUR"]:
            new_color = 3
        elif new_color.upper() in ["BGRA", "TRANSPARENT"]:
            new_color = 4
        else:
            new_color = -1

    if new_color not in [1, 3, 4]:
        raise ValueError(
            "The image type {} is not supported. Currently supported: GRAY (1), BGR (3) and BGRA (4)".format(
                new_color
            )
        )

    # Determine the color type of current image
    if len(img.shape) == 2:
        currentType = 1
    elif len(img.shape) != 3:
        raise ValueError(
            "The image does not have known dimensions ({}). It should have 2 or 3 dimensions.".format(
                img.shape
            )
        )
    elif img.shape[2] == 1:
        img = img[:, :, 0]
        currentType = 1
    elif img.shape[2] == 3:
        currentType = 3
    elif img.shape[2] == 4:
        currentType = 4
    else:
        raise ValueError(
            "The image type {} is not supported. Currently supported: GRAY (1), BGR (3) and BGRA (4)".format(
                img.shape[2]
            )
        )

    # That was easy
    if currentType == new_color:
        return img

    if currentType == 1 and new_color == 3:
        return cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    if currentType == 1 and new_color == 4:
        return cv2.cvtColor(img, cv2.COLOR_GRAY2BGRA)

    if currentType == 3 and new_color == 1:
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    if currentType == 3 and new_color == 4:
        return cv2.cvtColor(img, cv2.COLOR_BGR2BGRA)

    if currentType == 4 and new_color == 1:
        return cv2.cvtColor(img, cv2.COLOR_BGRA2GRAY)
    if currentType == 4 and new_color == 3:
        return cv2.cvtColor(img, cv2.COLOR_BGRA2BGR)



from typing import Tuple

import cv2
import numpy as np


def get_center(contour: np.ndarray) -> Tuple[int, int]:
    """
    Returns the center of a contour

    Args:
        contour (np.ndarray): An opencv contour of a single object

    Returns:
        x: (float) Center x postion
        y: (float) Center y postion

    """
    M = cv2.moments(contour)
    if M["m00"] == 0:
        return contour[0][0][0], contour[0][0][1]
    cX = int(M["m10"] / M["m00"])
    cY = int(M["m01"] / M["m00"])

    return cX, cY



import cv2
import numpy as np


def circularity(contour: np.ndarray) -> float:
    """
    Calculates the circularity of a contour.
    This is based on the perimeter and area of the contour.
    A circle will give the smallest area for a given perimeter, so this is 1.

    The smaller the area is for a perimeter the less circular the contour is.
    So the closer is get to 0.

    The formula works from calculating the r**2 based on Area and Perimeter
    Area based:
    r**2 = A/pi

    Perimeter based:
    r**2 = (p/2*pi)**2

    Dividing these two gives circularity
    c = (A/pi) / (p/2*pi)**2 = (4 * pi * A) / (p**2)

    Args:
        contour (np.ndarray): Contour as given by opencv

    Returns:
        float: circularity (0 < c <= 1)
    """
    # calculate size
    area = cv2.contourArea(contour)

    # calculate shape
    perimeter = cv2.arcLength(contour, True)

    circularity = (4 * np.pi * area) / (perimeter**2)

    return circularity



import numpy as np
import cv2
from typing import Union


def sendErrorShape(shapeC):
    raise TypeError("Expect contour in form (N, 2) not {}".format(shapeC))


def make_mask_contour(img_shape: tuple, contour: Union[list, np.ndarray]) -> np.ndarray:
    """
    This function produces a boolean image with the contour given.

    img_shape: (tuple, shape (2,) ) the shape of the image given in image coordinates.
        These can be obtained with img_dim.
    contour: (list, np.ndarray) This is a contour as outputted by opencv

    return:
        numpy array with type boolean
    """
    contour = np.array(contour, dtype=np.int32)
    shapeC = np.shape(contour)

    if len(shapeC) != 2:
        sendErrorShape(shapeC)

    if shapeC[1] != 2:
        sendErrorShape(shapeC)

    contour = contour.reshape((1, shapeC[0], 2))

    mask = np.zeros((img_shape[1], img_shape[0]), dtype=np.uint8)
    cv2.fillPoly(mask, contour, 1)
    return mask



import numpy as np

from ..labels2listsOfPoints import labels2listsOfPoints
from ..positions2contour import positions2contour


def labels2contours(
    labels: np.ndarray, simplify_error: float = 0, return_inner_contours: bool = False
) -> list:
    """
    Transforms an image with labels into opencv contours

    labels: (numpy.array) An array where every pixels is label to which object it belongs
    simplify_error ⚠️: (float) DEPRECATED the amount of error allowed well simplifying the contours
    return_inner_contours (bool, optional): return the internal contours.
            These contours are around the holes with the contour
            default: False

    """
    listPoints = labels2listsOfPoints(labels)
    contours = [
        positions2contour(
            p,
            simplify_error=simplify_error,
            return_inner_contours=return_inner_contours,
        )
        for p in listPoints
        if len(p) > 0
    ]
    return contours



from unittest import TestCase
import numpy as np
from .main import get_center


class TestGetCenterContour(TestCase):
    def test_happy_flow(self):
        contour = np.array([[[500, 50]], [[500, 99]], [[539, 99]], [[539, 50]]])

        expectedResult = (519, 74)

        result = get_center(contour=contour)

        self.assertEqual(len(result), len(expectedResult))
        for i in range(len(result)):
            self.assertEqual(result[i], expectedResult[i])

    def test_single_point(self):
        contour = np.array([[[500, 50]]])

        expectedResult = (500, 50)

        result = get_center(contour=contour)

        self.assertEqual(len(result), len(expectedResult))
        for i in range(len(result)):
            self.assertEqual(result[i], expectedResult[i])



import copy
from typing import Union


def rotate_json(json_object: dict, angle: int, img_shape: Union[list, tuple]) -> dict:
    """
    Takes in a json object (ellipse or polygon, standard cytosmart fromat) and rotates it.
    Possible rotations are 0, 90, 180 and 270 degree.
    json_object: (dict) following the standard cytosmart format
    angle: (int) the angle of rotations. Can chose from 0, 90, 180, 270 degree.
    img_shape: (list or tuple) the shape of the image the json_object is related to.

    example:
    input_json = {
            "type": "ellipse",
            "center": {"x": 0, "y": 2},
            "radiusX": 5,
            "radiusY": 5,
            "angleOfRotation": 0,
        }
    angle = 90

    expected_json = {
            "type": "ellipse",
            "center": {"x": 4, "y": 3},
            "radiusX": 5,
            "radiusY": 5,
            "angleOfRotation": 0,
        }
    """
    assert len(img_shape) == 2

    newjson_object = copy.deepcopy(json_object)
    if json_object["type"] == "ellipse":
        if angle == 0:
            newjson_object["center"]["x"] = json_object["center"]["x"]
            newjson_object["center"]["y"] = json_object["center"]["y"]
        elif angle == 90:
            newjson_object["center"]["x"] = (
                img_shape[0] - json_object["center"]["y"] - 1
            )
            newjson_object["center"]["y"] = json_object["center"]["x"]

        elif angle == 180:
            newjson_object["center"]["x"] = (
                img_shape[1] - json_object["center"]["x"] - 1
            )
            newjson_object["center"]["y"] = (
                img_shape[0] - json_object["center"]["y"] - 1
            )

        elif angle == 270:
            newjson_object["center"]["x"] = json_object["center"]["y"]
            newjson_object["center"]["y"] = (
                img_shape[1] - json_object["center"]["x"] - 1
            )

        else:
            raise ValueError("Unkown Angle should be 0, 90, 180 or 270")
    elif json_object["type"] == "polygon":
        if angle == 0:
            newjson_object["points"] = json_object["points"]

        elif angle == 90:
            for point_n in range(len(newjson_object["points"])):
                newjson_object["points"][point_n]["x"] = (
                    img_shape[0] - json_object["points"][point_n]["y"] - 1
                )
                newjson_object["points"][point_n]["y"] = json_object["points"][point_n][
                    "x"
                ]

        elif angle == 180:
            for point_n in range(len(newjson_object["points"])):
                newjson_object["points"][point_n]["x"] = (
                    img_shape[1] - json_object["points"][point_n]["x"] - 1
                )
                newjson_object["points"][point_n]["y"] = (
                    img_shape[0] - json_object["points"][point_n]["y"] - 1
                )

        elif angle == 270:
            for point_n in range(len(newjson_object["points"])):
                newjson_object["points"][point_n]["x"] = json_object["points"][point_n][
                    "y"
                ]
                newjson_object["points"][point_n]["y"] = (
                    img_shape[1] - json_object["points"][point_n]["x"] - 1
                )

        else:
            raise ValueError("Unkown Angle should be 0, 90, 180 or 270")

    return newjson_object



from unittest import TestCase
import numpy as np
from numpy import array
from .main import convert_color


class Testconvert_color(TestCase):
    def test_gray2grey(self):
        inputArr = np.zeros((30, 40), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40), dtype=np.uint8)
        expected[3, 4] = 40

        result = convert_color(inputArr, "grey")

        np.testing.assert_array_equal(result, expected)

    def test_BGR2grey(self):
        inputArr = np.zeros((30, 40, 3), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40), dtype=np.uint8)
        expected[3, 4] = 40

        result = convert_color(inputArr, 1)

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_BGR2grey_no_int_type(self):
        inputArr = np.zeros((30, 40, 3), dtype=np.float16)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40), dtype=np.uint8)
        expected[3, 4] = 40

        result = convert_color(inputArr, 1)

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_BGRa2grey(self):
        inputArr = np.zeros((30, 40, 4), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40), dtype=np.uint8)
        expected[3, 4] = 40

        result = convert_color(inputArr, "gray")

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_gray2BGR(self):
        inputArr = np.zeros((30, 40), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40, 3), dtype=np.uint8)
        expected[3, 4] = 40

        result = convert_color(inputArr, 3)

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_gray2BGR_3dimension_input(self):
        inputArr = np.zeros((30, 40, 1), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40, 3), dtype=np.uint8)
        expected[3, 4] = 40

        result = convert_color(inputArr, "colour")

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_BGRa2BGR(self):
        inputArr = np.zeros((30, 40, 4), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40, 3), dtype=np.uint8)
        expected[3, 4] = 40

        result = convert_color(inputArr, "BGr")

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_gray2BGRA(self):
        inputArr = np.zeros((30, 40), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40, 4), dtype=np.uint8)
        expected[3, 4] = 40
        expected[:, :, 3] = 255

        result = convert_color(inputArr, 4)

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_BGR2BGRA(self):
        inputArr = np.zeros((30, 40, 3), dtype=np.uint8)
        inputArr[3, 4] = 40

        expected = np.zeros((30, 40, 4), dtype=np.uint8)
        expected[3, 4] = 40
        expected[:, :, 3] = 255

        result = convert_color(inputArr, "TranspaREnt")

        print(result)
        np.testing.assert_array_equal(result, expected)

    def test_unkownInputColor(self):
        inputArr = np.zeros((30, 40, 7), dtype=np.uint8)
        inputArr[3, 4] = 40

        self.assertRaises(ValueError, convert_color, inputArr, 3)

    def test_unkownInputShape(self):
        inputArr = np.zeros((30, 40, 3, 2), dtype=np.uint8)
        inputArr[3, 4] = 40

        self.assertRaises(ValueError, convert_color, inputArr, 3)

    def test_unkownOutputColorStr(self):
        inputArr = np.zeros((30, 40, 3), dtype=np.uint8)
        inputArr[3, 4] = 40

        self.assertRaises(ValueError, convert_color, inputArr, "Unicorn")

    def test_unkownOutputColorInt(self):
        inputArr = np.zeros((30, 40, 3), dtype=np.uint8)
        inputArr[3, 4] = 40

        self.assertRaises(ValueError, convert_color, inputArr, 42)



import math

import cv2

from ...contour_operations import circularity
from ...transformers import json2contours


def add_circularity(json_object: dict) -> None:
    """
    This function adds the property 'circularity' to the current json object.
    circularity is calculated by drawing a circle around the object,
    and then compare the area of the circle with area of the object.
    The bigger the difference between those 2 areas, the lower the circularity.

    json_object: (dict) a json object of type polygon
    """
    if json_object["type"] == "polygon":
        cnt = json2contours(json_object)

        _circularity = circularity(cnt)

        # Add results to current object
        json_object["circularity"] = _circularity
    else:
        raise ValueError(
            f"Currently only the polygon objects are supported; not {json_object['type']}"
        )



from unittest import TestCase

from tomni.annotation_manager.annotations.point import Point
from tomni.annotation_manager.annotations.polygon import Polygon


class TestPolygonCompare(TestCase):
    def test_positive_squares(self):
        poly1 = Polygon(
            [Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3), Point(2, 1)],
            id=1,
            label="square",
        )
        poly2 = Polygon(
            [Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3), Point(2, 1)],
            id=2,
            label="square",
        )

        self.assertTrue(poly1 == poly2)

    def test_flipped_squares(self):
        poly1 = Polygon(
            [Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3), Point(2, 1)],
            id=1,
            label="square",
        )
        poly2 = Polygon(
            [Point(2, 1), Point(2, 3), Point(1, 3), Point(1, 2), Point(1, 1)],
            id=2,
            label="flipped square",
        )

        self.assertTrue(poly1 == poly2)

    def test_rotated_squares(self):
        poly1 = Polygon(
            [Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3), Point(2, 1)],
            id=1,
            label="square",
        )
        poly2 = Polygon(
            [Point(2, 1), Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3)],
            id=2,
            label="rotated square",
        )

        self.assertTrue(poly1 == poly2)

    def test_rotated_and_flipped_squares(self):
        poly1 = Polygon(
            [Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3), Point(2, 1)],
            id=1,
            label="square",
        )
        poly2 = Polygon(
            [Point(2, 3), Point(1, 3), Point(1, 2), Point(1, 1), Point(2, 1)],
            id=2,
            label="flipped rotated square",
        )

        self.assertTrue(poly1 == poly2)

    def test_positive_squares_hourglass(self):
        poly1 = Polygon(
            [Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3), Point(2, 1)],
            id=1,
            label="square",
        )
        poly2 = Polygon(
            [Point(1, 1), Point(2, 3), Point(1, 3), Point(1, 2), Point(2, 1)],
            id=2,
            label="hourglass",
        )

        self.assertFalse(poly1 == poly2)

    def test_positive_wrong_type(self):
        poly1 = Polygon(
            [Point(1, 1), Point(1, 2), Point(1, 3), Point(2, 3), Point(2, 1)],
            id=1,
            label="square",
        )
        point2 = Point(1, 1)

        self.assertFalse(poly1 == point2)



# %%
from typing import Tuple, Union

from shapely import affinity
from shapely.geometry import MultiPolygon, Polygon, point
from shapely.ops import polygonize


def get_boundingbox(annotation_object: dict) -> Tuple[int, int, int, int]:
    """Calculate the bounding box of the dict annotation object.

    Args:
        object (dict): A JSON annotation object.

    Raises:
        ValueError: If object type is not known.

    Returns:
        Tuple[int,int,int,int]: A tuple containing the min_x, min_y, max_x and max_y of the bounding box.
    """
    if annotation_object["type"] == "polygon":
        all_x = [j["x"] for j in annotation_object["points"]]
        all_y = [j["y"] for j in annotation_object["points"]]
        return (min(all_x), min(all_y), max(all_x), max(all_y))
    elif annotation_object["type"] == "ellipse":
        return (
            annotation_object["center"]["x"] - annotation_object["radiusX"],
            annotation_object["center"]["y"] - annotation_object["radiusY"],
            annotation_object["center"]["x"] + annotation_object["radiusX"],
            annotation_object["center"]["y"] + annotation_object["radiusY"],
        )
    else:
        raise ValueError(f"Object 1 has an unkown type of {annotation_object['type']}")


def are_boundingboxes_overlapping(
    bbox1: Tuple[int, int, int, int], bbox2: Tuple[int, int, int, int]
) -> bool:
    """Checks if bounding boxes are overlapping and returns truth value.

    Args:
        bbox1 (Tuple[int,int,int,int]): A tuple containing the min_x, min_y, max_x and max_y of bounding box 1.
        bbox2 (Tuple[int,int,int,int]): A tuple containing the min_x, min_y, max_x and max_y of bounding box 2.

    Returns:
        bool: Returns a truth value indicating whether the bounding boxes are overlapping.
    """
    if bbox1[0] >= bbox2[2]:
        return False
    elif bbox2[0] >= bbox1[2]:
        return False
    elif bbox1[1] >= bbox2[3]:
        return False
    elif bbox1[3] <= bbox2[1]:
        return False
    else:
        return True


def create_ellipse(center: Tuple[int, int], radii: Tuple[int, int], angle: Union[int, None] = 0):
    if radii[0] < 0 or radii[1] < 0:
        raise ValueError("Radi cannot be negative")
    circ = point.Point(center).buffer(1)
    ell = affinity.scale(circ, int(radii[0]), int(radii[1]))
    ellr = affinity.rotate(ell, angle)

    return ellr


def object2shape(annotation_object: dict):
    """Convert object dict to shapely shape.

    Args:
        object (dict): An annotation object in cytosmart format.

    Raises:
        ValueError: If type of the annotation object is unknown.

    Returns:
        Returns a shapely shape.
    """
    if annotation_object["type"] == "polygon":
        object_points = [[j["x"], j["y"]] for j in annotation_object["points"]]
        if len(object_points) < 3:
            return Polygon()

        polygon = Polygon(object_points)
        if polygon.is_valid:
            return polygon
        else:
            be = polygon.exterior
            mls = be.intersection(be)
            polygons = polygonize(mls)
            multi_polygons = MultiPolygon(polygons)
            return multi_polygons

    elif annotation_object["type"] == "ellipse":
        return create_ellipse(
            (annotation_object["center"]["x"], annotation_object["center"]["y"]),
            (annotation_object["radiusX"], annotation_object["radiusY"]),
            annotation_object["angleOfRotation"],
        )
    else:
        raise ValueError(f"Object 1 has an unkown type of {annotation_object['type']}")


def overlap_object(object_1: dict, object_2: dict) -> float:
    """This function calculates the overlap percentage of the first object with the second object.

    Args:
        object1 (dict): JSON object1 in cytosmart format
        object2 (dict): JSON object2 in cytosmart format

    Returns:
        int: overlap percentage.
    """

    bbox1 = get_boundingbox(object_1)
    bbox2 = get_boundingbox(object_2)

    if not are_boundingboxes_overlapping(bbox1, bbox2):
        return 0

    shape1 = object2shape(object_1)
    shape2 = object2shape(object_2)

    intersect = shape1.intersection(shape2)
    overlap_percentage = intersect.area / (shape1.area + 1e-17)
    return overlap_percentage



from ...transformers import json2contours
from ...contour_operations import get_center


def add_center(json_object: dict):

    if json_object["type"] == "ellipse":
        pass
    elif json_object["type"] == "polygon":
        contour = json2contours(json_object)
        x, y = get_center(contour)
        json_object["center"] = {"x": x, "y": y}
    else:
        raise ValueError(f"No json type found of {json_object['type']}")



from unittest import TestCase
import numpy as np
from .main import contours2json


class TestContour2json(TestCase):
    def test_happy_flow(self):
        inputContours = np.array([[[[500, 50]], [[500, 99]], [[539, 99]], [[539, 50]]]])

        expectedResult = [
            {
                "type": "polygon",
                "points": [
                    {"x": 500, "y": 50},
                    {"x": 500, "y": 99},
                    {"x": 539, "y": 99},
                    {"x": 539, "y": 50},
                ],
                "id": "unicorn",
                "parents": [],
                "children": [],
            }
        ]

        result = contours2json(inputContours)
        result[0]["id"] = "unicorn"

        self.assertEqual(len(result), len(expectedResult))
        for i in range(len(result)):
            print(
                f"future {i}: \n input \n {result[i]} \n expected \n {expectedResult[i]}"
            )
            self.assertDictEqual(result[i], expectedResult[i])



from math import pi, sqrt
from typing import Tuple

import cv2
import numpy as np

from .. import get_center


def approximate_circle_by_area(contour: np.ndarray) -> Tuple[float, float, float]:
    """
    Returns a circle with the same area and center as the contour.

    Args:
        contour (np.ndarray): An opencv contour of a single object

    Returns:
        x: (float) Center x postion
        y: (float) Center y postion
        radius: (float) Radius of the circle
    """
    x, y = get_center(contour)
    area = cv2.contourArea(contour)
    radius = sqrt(area / pi)
    return x, y, radius



from unittest import TestCase

from tomni.annotation_manager import Ellipse, Point


class TestEllipseCompare(TestCase):
    def test_ignore_decimals(self):
        ellip1 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=0, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=1.0,
            radius_y=3.0,
            center=Point(10, 40),
            rotation=0.0,
            id="ellipse2",
        )

        self.assertTrue(ellip1 == ellip2)

    def test_different_radius_x(self):
        ellip1 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=45, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=2, radius_y=3, center=Point(10, 40), rotation=45, id="ellipse2"
        )

        self.assertFalse(ellip1 == ellip2)

    def test_different_radius_y(self):
        ellip1 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=100, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=1, radius_y=2, center=Point(10, 40), rotation=100, id="ellipse2"
        )

        self.assertFalse(ellip1 == ellip2)

    def test_different_center(self):
        ellip1 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=-100, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=1, radius_y=3, center=Point(11, 40), rotation=-100, id="ellipse2"
        )

        self.assertFalse(ellip1 == ellip2)

    def test_different_rotation(self):
        ellip1 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=0, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=10, id="ellipse2"
        )

        self.assertFalse(ellip1 == ellip2)

    def test_rotation_90(self):
        ellip1 = Ellipse(
            radius_x=3, radius_y=1, center=Point(10, 40), rotation=90, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=0, id="ellipse2"
        )

        self.assertTrue(ellip1 == ellip2)

    def test_rotation_180(self):
        ellip1 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=200, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=20, id="ellipse2"
        )

        self.assertTrue(ellip1 == ellip2)

    def test_negative_rotation(self):
        ellip1 = Ellipse(
            radius_x=1, radius_y=3, center=Point(10, 40), rotation=-10, id="ellipse1"
        )
        ellip2 = Ellipse(
            radius_x=3, radius_y=1, center=Point(10, 40), rotation=80, id="ellipse2"
        )

        self.assertTrue(ellip1 == ellip2)



import uuid
from typing import Union


def ellipse2json(
    x: int, y: int, radius: int, secondRadius: Union[None, int] = None, alpha: float = 0
) -> dict:
    """
    x: (int) center x position
    y: (int) center y position
    radius: (int) the radius of the circle
    secondRadius: (int, None) (optional) In case of an ellipse, this is the smallest radius of the two.
    alpha: (float) (optional) In case of an ellipse. This is the angle between the biggest radius and the x-axis.
    """
    result = []
    result = {
        "type": "ellipse",
        "center": {"x": int(x), "y": int(y)},
        "radiusX": int(radius),
        "radiusY": int(secondRadius) if secondRadius else int(radius),
        "angleOfRotation": alpha,
        "id": str(uuid.uuid4()),
    }
    return result



import numpy as np


def fit_rect_around_ellipse(
    x: int, y: int, r1: float, r2: float, alpha: float
) -> (int, int, int, int):
    """
    This function gives the boundingbox around an ellipse.
    x, y: (int, int) middle point of ellipse
    r1: (float) longest radius
    r2: (float) shortest radius. By definition 90 degrees away from r1
    alpha: (float) degrees between r1 and x-axis.
    Special thanks to: https://math.stackexchange.com/questions/91132/how-to-get-the-limits-of-rotated-ellipse
    """
    alpha = alpha / 180 * np.pi
    a = np.sqrt(r1**2 * np.cos(alpha) ** 2 + r2**2 * np.sin(alpha) ** 2)
    b = np.sqrt(r1**2 * np.sin(alpha) ** 2 + r2**2 * np.cos(alpha) ** 2)
    x1 = int(round(x - a))
    y1 = int(round(y - b))
    x2 = int(round(x + a))
    y2 = int(round(y + b))
    return x1, y1, x2, y2



import numpy as np
import cv2


def positions2contour(
    positions: np.ndarray,
    simplify_error: float = 0,
    return_inner_contours: bool = False,
) -> np.ndarray:
    """Transforms a list of positions to opencv contour format

    Args:
        positions (np.ndarray): 2 dimensional array with shape [N, 2].
            For the N pixels of the shape given as x, y
        simplify_error ⚠️ (float, optional): Deprecated!!
        return_inner_contours (bool, optional): return the internal contours.
            These contours are around the holes with the contour
            default: False

    Raises:
        DeprecationWarning: simplification is no longer supported

    Returns:
        np.ndarray: Opencv contour
    """
    if simplify_error > 0:
        raise DeprecationWarning("As of Tomni 1.10 simplify is no longer supported")

    max_x = np.max(positions[:, 0])
    min_x = np.min(positions[:, 0])
    max_y = np.max(positions[:, 1])
    min_y = np.min(positions[:, 1])

    positions -= np.array([min_x, min_y])
    transposed_positions = positions.transpose()

    array_mask = np.zeros((max_y + 1 - min_y, max_x + 1 - min_x), dtype="uint8")
    array_mask[transposed_positions[1], transposed_positions[0]] = 255

    # Chose between RETR_CCOMP to retrieve all of the contours and organizes them into a two-level hierarchy
    # Or RETR_EXTERNAL retrieves only the extreme outer contours
    mode = cv2.RETR_CCOMP if return_inner_contours else cv2.RETR_EXTERNAL
    contours = cv2.findContours(array_mask, mode, cv2.CHAIN_APPROX_SIMPLE)[0]

    external_contour = contours[0] + [min_x, min_y]

    if not return_inner_contours:
        return external_contour

    if len(contours) == 1:
        internal_contours = []
    else:
        internal_contours = [i + [min_x, min_y] for i in contours[1:]]

    return external_contour, internal_contours



from __future__ import absolute_import
from unittest import TestCase
from .main import make_mask_circle
import numpy as np


class TestCircle_mask_selector(TestCase):
    def test_5x5_5(self):
        expected = np.array(
            [
                [0, 1, 1, 1, 0],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [0, 1, 1, 1, 0],
            ]
        )

        result = make_mask_circle((5, 5), 5)
        np.testing.assert_array_equal(result, expected)

    def test_5x7_5(self):
        expected = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 1, 1, 0],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [0, 1, 1, 1, 0],
                [0, 0, 0, 0, 0],
            ]
        )

        result = make_mask_circle((5, 7), 5)
        np.testing.assert_array_equal(result, expected)

    def test_7x7_5(self):
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ]
        )

        result = make_mask_circle((7, 7), 5)
        np.testing.assert_array_equal(result, expected)

    def test_7x10_6(self):
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ]
        )

        result = make_mask_circle((7, 10), 6)
        np.testing.assert_array_equal(result, expected)

    def test_big_301x201(self):
        result = make_mask_circle((301, 201), 101)
        self.assertEqual(np.shape(result), (201, 301))
        self.assertEqual(result[0, 0], False)
        self.assertEqual(result[0, 150], False)
        self.assertEqual(result[25, 75], False)
        self.assertEqual(result[100, 150], True)
        self.assertEqual(result[63, 117], True)

    def test_overflow_5001(self):
        result = make_mask_circle((5001, 5001), 5000)

        self.assertEqual(result[0, 0], False)
        self.assertEqual(result[5000, 5000], False)
        self.assertEqual(result[4300, 4300], False)
        self.assertEqual(result[2500, 2500], True)
        self.assertEqual(result[3300, 3300], True)

    # # RUN THIS LOCALY. This is to big for the build agent
    # def test_overflow_25001(self):
    #     result = make_mask_circle((25001, 25001),25000)
    #
    #     self.assertEqual(result[0, 0], False)
    #     self.assertEqual(result[25000, 25000], False)
    #     self.assertEqual(result[21400, 21400], False)
    #     self.assertEqual(result[21300, 21300], True)
    #     self.assertEqual(result[12500, 12500], True)



import cv2
import numpy as np


def bbox_fitting(
    img: np.ndarray, x1: int, y1: int, x2: int, y2: int, padding_value: int = 0
) -> np.ndarray:
    """
    Bounding box fitting:
    Using cropping and padding to fit an image to the bounding box.
    Relative to the 0,0 of the input image the image gets cropped and padded to fit with x1, y1, x2 and y2.
    The result is including the pixels x1, x2, y1 and y2.
    So a 0, 0, 9, 9 gives an 10x10 result.

    WARNING:
    Coordinates are IMAGE-coordinates. So that is the same pillow and OpenCV uses but swapped from what Numpy and Scipy use (ARRAY-coordinates).
    Yes, I hate it too.

    Args:
        img: (np.ndarray): The image the bounding box will be extracted from

        x1: (int): Lowest value of X

        y1: (int): Lowest value of Y

        x2: (int): Highest value of X

        y2: (int): Highest value of Y

        padding_value: (int, optional): Every pixel outside the image bit inside the bounding box will get this value.
        Defaults to 0.

    Returns:
        np.ndarray: Extracted patch within the bounding box from the image
    """
    """

    """

    if not isinstance(img, np.ndarray):
        raise TypeError("Img needs to be a numpy.ndarry not {}".format(type(img)))

    if (
        not isinstance(x1, int)
        or not isinstance(x2, int)
        or not isinstance(y1, int)
        or not isinstance(y2, int)
    ):
        raise ValueError(
            "x1, y1, x2 and y2 need to be a positive int not {}, {}, {}, {}".format(
                type(x1), type(y1), type(x2), type(y2)
            )
        )

    if x2 < 0:
        return np.zeros((y2 - y1, x2 - x1))
    if y2 < 0:
        return np.zeros((y2 - y1, x2 - x1))
    if x1 >= img.shape[1]:
        return np.zeros((y2 - y1, x2 - x1))
    if y1 >= img.shape[0]:
        return np.zeros((y2 - y1, x2 - x1))

    cy1 = max(0, y1)
    cy2 = min(img.shape[0], y2)
    cx1 = max(0, x1)
    cx2 = min(img.shape[1], x2)

    pad_left = -min(0, x1)
    pad_right = max(x2 - img.shape[1], 0)
    pad_top = -min(0, y1)
    pad_bottom = max(y2 - img.shape[0], 0)

    img = img[cy1:cy2, cx1:cx2]

    img = cv2.copyMakeBorder(
        img,
        top=pad_top,
        bottom=pad_bottom,
        left=pad_left,
        right=pad_right,
        borderType=cv2.BORDER_CONSTANT,
        value=padding_value,
    )

    return img



import numpy as np


def img_paste(big_img, small_img, x, y):
    """
    This function adds a small image to a bigger image by translating it to position x, y (img coordinates).
    :big_img: (numpy.ndarray) the big image that gets altered
    :small_img: (numpy.ndarray) the smaller image that gets added to the big_img
    :x: (int) image coordinate x
    :y: (int) image coordinate y

    return: None (big_img gets altered)
    """
    size_bimg = (big_img.shape[1], big_img.shape[0])
    size_simg = (small_img.shape[1], small_img.shape[0])

    # calculate position in big image
    # START
    if x < 0:
        start_x = 0
        crop_x_start = -x
    elif x > size_bimg[0] - 1:
        return
    else:
        start_x = x
        crop_x_start = 0

    if y < 0:
        start_y = 0
        crop_y_start = -y
    elif y > size_bimg[1] - 1:
        return
    else:
        start_y = y
        crop_y_start = 0

    # STOP
    if x + size_simg[0] > size_bimg[0]:
        stop_x = size_bimg[0]
        if x >= 0:
            crop_x_stop = size_bimg[0] - x
        else:
            crop_x_stop = size_bimg[0] + crop_x_start

    elif x + size_simg[0] < 0:
        return
    else:
        stop_x = x + size_simg[0]
        crop_x_stop = size_simg[0]

    if y + size_simg[1] > size_bimg[1]:
        stop_y = size_bimg[1]
        if y >= 0:
            crop_y_stop = size_bimg[1] - y
        else:
            crop_y_stop = size_bimg[1] + crop_y_start
    elif y + size_simg[1] < 0:
        return
    else:
        stop_y = y + size_simg[1]
        crop_y_stop = size_simg[1]

    # Crop smaller image to fit
    overlay_img = small_img[crop_y_start:crop_y_stop, crop_x_start:crop_x_stop]
    big_img[start_y:stop_y, start_x:stop_x] = overlay_img



import numpy as np
from scipy.sparse import csr_matrix, find
import numpy_indexed as npi


def labels2listsOfPoints(labels: np.ndarray) -> np.ndarray:
    """
    Transforms each label into a list of points.
    These points are for every point in the label.

    labels: (numpy.array) An array where every pixel is labeled to which object it belongs
    """
    if np.max(labels) < 256:
        dataType = np.uint8
    elif np.max(labels) < 65536:
        dataType = np.uint16
    elif np.max(labels) < 4294967296:
        dataType = np.uint32
    else:
        raise ValueError("There are to many labels")

    # Memory check number of labels
    c = csr_matrix(labels, dtype=dataType)
    x, y, v = find(c)
    g = npi.group_by(v)
    output = g.split_array_as_list(np.array([y, x]).transpose())
    return output



from unittest import TestCase

from tomni.annotation_manager.annotations.point import Point


class TestPointCompare(TestCase):
    def test_positive(self):
        point1 = Point(1, 2)
        point2 = Point(1, 2)

        self.assertTrue(point1 == point2)

    def test_positive_decimal(self):
        point1 = Point(1, 2)
        point2 = Point(1.0, 2.0)

        self.assertTrue(point1 == point2)

    def test_negative(self):
        point1 = Point(1, 2)
        point2 = Point(2, 1)

        self.assertFalse(point1 == point2)



from typing import List


def json2dict(
    json_list: List[dict], keywords: list = ["type", "area", "center", "id"]
) -> dict:
    """Transforms a list of standard CytoSMART format jsons to a dict using a list of given keywords.
       This dict contains list of each keyword extracted from the json objects and a list with indexes.
       e.g json_dict{"area": [100, 33,44], 'type': ["ellipse", "polygon", "ellipse"], "index":[0,1,2]}

    Args:
        json_list (List[dict]):the jsons are in the standard CytoSMART format.
        keywords (list, optional): a list with keywords you want to extract from the json_objects.
                Defaults to ["type", "area", "center", "id"].

    Returns:
        dict: a dict containing list of each keyword extracted from the json objects
    """
    keywords.append("index")
    json_dict = {}

    for key in keywords:
        json_dict[key] = []

    for i, json_object in enumerate(json_list):
        for key in keywords:

            if key == "index":
                json_dict[key].append(i)
            elif key not in json_object.keys() and key != "index":
                json_dict[key].append(None)
            else:
                if isinstance(json_object[key], dict):
                    variable = [json_object[key][x] for x in json_object[key].keys()]
                else:
                    variable = json_object[key]
                json_dict[key].append(variable)

    return json_dict



from __future__ import division
import numpy as np
import cv2
from . import bbox_fitting


def make_mask_ellipse(image_size, x1, y1, rx, ry):

    """
    :param image_size: the size of the image
    :param x1: x coordinate of the center of the ellipse
    :param y1: y coordinate of the center of the ellipse
    :param rx: the length of the radius on the x axis of the ellipse
    :param ry: the length of the radius on the y axis of the ellipse
    :return: an image with an ellipse of True surrounded by False

    The big ellipse function is less precise but better at memory management.
    The small ellipse function is more precise.

    """
    x1 = int(x1)
    y1 = int(y1)
    rx = int(rx)
    ry = int(ry)

    if rx < 1 or ry < 1:
        raise ValueError("Radii must be greater than 1.")
    else:
        if (rx < 100) and (ry < 100):
            return make_small_mask_ellipse(image_size, x1, y1, rx, ry)
        else:
            return make_big_mask_ellipse(image_size, x1, y1, rx, ry)


def make_small_mask_ellipse(image_size, x1, y1, rx, ry):
    """

    :param image_size: the size of the image
    :param x1: x coordinate of the center of the ellipse
    :param y1: y coordinate of the center of the ellipse
    :param rx: the length of the radius on the x axis of the ellipse
    :param ry: the length of the radius on the y axis of the ellipse
    :return: an image with an ellipse of True surrounded by False

    Futere request: Variable Alpha.

    For rotated ellipses, the alpha value is used to find the translated points
    alpha = angle of rotation (in degrees)
    alpha = 0 gives no rotation of the ellipse

    The sin() and cos() values are calculated using math library

    The following formula can be used to calculate the coordinates:
    circle = ((xx * math.cos(alpha) + yy * math.sin(alpha)) ** 2) * (rx ** 2) +\
             # ((xx * math.sin(alpha) - yy * math.cos(alpha)) ** 2) * (ry ** 2)
    """

    yy, xx = np.mgrid[: image_size[1], : image_size[0]]

    xx = xx.astype(np.int32)
    yy = yy.astype(np.int32)

    # Center point of the ellipse becomes the (0, 0) point of the image
    xx -= x1
    yy -= y1

    # The ellipse formula used to calculate the points is
    # x^2 * (ry)^2 + y^2 * (rx)^2
    # Calculating the value of each point depending on the ellipse equation
    ellipse = (xx**2) * (ry**2) + (yy**2) * (rx**2)

    # The ellipse formula determines if a point is inside the ellipse or not
    # x^2 * (ry)^2 + y^2 * (rx)^2 <= (rx)^2 * (ry)^2
    kernel = ellipse <= (rx**2) * (ry**2)
    kernel = kernel * (np.abs(xx) <= rx)
    kernel = kernel * (np.abs(yy) <= ry)
    return kernel


def make_big_mask_ellipse(image_size, xe, ye, rex, rey):
    """

    :param image_size:
    :param xe: x coordinate of the ellipse
    :param ye: y coordinate of the ellipse
    :param rex: the radius on the x axis of the ellipse
    :param rey: the radius on the y axis of the ellipse
    :return: an array with True and False corresponding to the ellipse mask in the image
    """

    out = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (rex * 2, rey * 2))
    x0 = -(xe - rex)
    y0 = -(ye - rey)
    x1 = image_size[0] + x0
    y1 = image_size[1] + y0
    out = bbox_fitting(out, x0, y0, x1, y1)
    return out



from __future__ import absolute_import, division
from typing import List
from . import bbox_fitting
from math import ceil
import numpy as np


def bbox_fitting_center(img: np.ndarray, size, padding_value: int = 0) -> np.ndarray:
    """
    Creates an image of size 'size' ([(int) x, (int) y]) in IMAGE-Coordinates).
    The orignal image will be centered and padded and/or cropped to fit the size.

    img: numpy.ndarray
    size: [int, int]

    output: numpy.ndarray
    """
    if not isinstance(img, np.ndarray):
        raise TypeError("Img needs to be a numpy.ndarry not {}".format(type(img)))

    if not isinstance(size[0], int) or not isinstance(size[1], int):
        raise ValueError(
            "x1, y1, x2 and y2 need to be a positive int not {}, {}".format(
                type(size[0]), type(size[1])
            )
        )
    x1 = -(size[0] - img.shape[1]) / 2
    y1 = -(size[1] - img.shape[0]) / 2
    x2 = x1 + size[0]
    y2 = y1 + size[1]

    # Round the number to fit exactly with the pixels
    x1 = int(ceil(x1))
    y1 = int(ceil(y1))
    x2 = int(ceil(x2))
    y2 = int(ceil(y2))

    return bbox_fitting(img, x1=x1, y1=y1, x2=x2, y2=y2, padding_value=padding_value)



from unittest import TestCase
import numpy as np
from .main import circularity


class TestCircularity(TestCase):
    def test_perfect_square(self):
        contour = np.array([[[500, 50]], [[500, 99]], [[549, 99]], [[549, 50]]])

        expected_result = np.pi / 4

        result = circularity(contour)

        self.assertAlmostEqual(result, expected_result)

    def test_octagon(self):
        # Octagon is rounder then a square
        # This is not perfect but okay
        contour = np.array(
            [
                [[500, 50]],
                [[490, 75]],
                [[500, 99]],
                [[525, 109]],
                [[549, 99]],
                [[559, 75]],
                [[549, 50]],
                [[525, 40]],
            ]
        )

        expected_result = 0.9479813831233634

        result = circularity(contour)

        self.assertAlmostEqual(result, expected_result)

    def test_flat_rectangle(self):
        # A rectangle of 450 by 50
        contour = np.array([[[0, 50]], [[0, 100]], [[450, 100]], [[450, 50]]])

        expected_result = 9 * np.pi / 100

        result = circularity(contour)

        self.assertAlmostEqual(result, expected_result)

    def test_square_with_dent(self):
        """
        A dent has a big impact on circularity because the perimeter get a lot bigger but the area gets smaller
        """
        contour = np.array(
            [
                [[500, 50]],
                [[500, 100]],
                # start dent
                [[524, 100]],
                [[524, 51]],
                [[525, 51]],
                [[525, 100]],
                # end ent
                [[550, 100]],
                [[550, 50]],
            ]
        )

        A = 50 * 50 - 49
        P = 50 * 4 + 2 * 49
        expected_result = (np.pi * 4 * A) / (P**2)  # 0.34683318742167396

        result = circularity(contour)

        self.assertAlmostEqual(result, expected_result, 5)



import cv2
import numpy as np

from typing import List


def binary2contours(
    binary_img: np.ndarray, return_inner_contours: bool = True
) -> List[np.ndarray]:
    """transforms a binary image to a list of outer contours and inner contours if boolean return_inner_contours is True.
    inner contours are in the same list as their corresponding outer contour.
    If return_inner_contours is False, only outer contours are returned.

    Args:
        binary_img (np.ndarray): A binary image.
        return_inner_contours (bool): boolean to check if inner contours are desired. Defaults to True.

    Returns:
        List[np.ndarray]: A list of outer contours with their corresponding inner contours.
    """

    mode = cv2.RETR_CCOMP if return_inner_contours else cv2.RETR_EXTERNAL

    contours, hierarchy = cv2.findContours(binary_img, mode, cv2.CHAIN_APPROX_SIMPLE)

    if not return_inner_contours:
        return contours

    combined_contours = []
    # Iterate over all contours and their hierarchies
    for idx, contour in enumerate(contours):
        current_hierarchy = hierarchy[0][idx]

        if current_hierarchy[-1] == -1:
            # If the contour has no parent, it is an outer contour
            combined_contour = [contour]
            inner_contours = []
            # Find the indices of the inner contours
            inner_indices = [i for i, h in enumerate(hierarchy[0]) if h[3] == idx]

            # Add the corresponding inner contours
            for inner_idx in inner_indices:
                inner_contours.append(contours[inner_idx])

            combined_contour.append(inner_contours)
            combined_contours.append(combined_contour)

    return combined_contours
